"""
PC-BASIC - compat.posix_console
POSIX console support with ANSI escape sequences

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

# pylint: disable=no-name-in-module

import os
import sys
import tty
import time
import termios
import select
import fcntl
import array
import struct
import atexit
from collections import deque

from .base import MACOS, PY2, HOME_DIR, wrap_input_stream, wrap_output_stream

if PY2:
    from .python2 import SimpleNamespace
else:
    from types import SimpleNamespace


# ANSI escape codes
# for reference, see:
# http://en.wikipedia.org/wiki/ANSI_escape_code
# http://misc.flogisoft.com/bash/tip_colors_and_formatting
# http://www.termsys.demon.co.uk/vtansi.htm
# http://ascii-table.com/ansi-escape-sequences-vt-100.php
# https://invisible-island.net/xterm/ctlseqs/ctlseqs.html


# ANSI escape sequences
ANSI = SimpleNamespace(
    RESET = u'\x1Bc',
    SET_SCROLL_SCREEN = u'\x1B[r',
    SET_SCROLL_REGION = u'\x1B[%i;%ir',
    RESIZE_TERM = u'\x1B[8;%i;%i;t',
    SET_TITLE = u'\x1B]2;%s\a',
    CLEAR_SCREEN = u'\x1B[2J',
    CLEAR_LINE = u'\x1B[2K',
    SCROLL_UP = u'\x1B[%iS',
    SCROLL_DOWN = u'\x1B[%iT',
    MOVE_CURSOR = u'\x1B[%i;%if',
    MOVE_RIGHT = u'\x1B[C',
    MOVE_LEFT = u'\x1B[D',
    MOVE_N_RIGHT = u'\x1B[%iC',
    MOVE_N_LEFT = u'\x1B[%iD',
    SHOW_CURSOR = u'\x1B[?25h',
    HIDE_CURSOR = u'\x1B[?25l',
    # 1 blinking block 2 block 3 blinking line 4 line
    SET_CURSOR_SHAPE = u'\x1B[%i q',
    SET_COLOUR = u'\x1B[%im',
    SET_CURSOR_COLOUR = u'\x1B]12;#%02x%02x%02x\a',
    SET_PALETTE_ENTRY = u'\x1B]4;%i;#%02x%02x%02x\a',
    RESET_PALETTE_ENTRY = u'\x1B]104;%i\a',
    #SAVE_CURSOR_POS = u'\x1B[s',
    #RESTORE_CURSOR_POS = u'\x1B[u',
    #REQUEST_SIZE = u'\x1B[18;t',
    #SET_FOREGROUND_RGB = u'\x1B[38;2;%i;%i;%im',
    #SET_BACKGROUND_RGB = u'\x1B[48;2;%i;%i;%im',
)

# ANSI base key codes
BASE_KEYS = dict(
    F1 = u'\x1B[11~',
    F2 = u'\x1B[12~',
    F3 = u'\x1B[13~',
    F4 = u'\x1B[14~',
    F5 = u'\x1B[15~',
    F6 = u'\x1B[17~',
    F7 = u'\x1B[18~',
    F8 = u'\x1B[19~',
    F9 = u'\x1B[20~',
    F10 = u'\x1B[21~',
    F11 = u'\x1B[23~',
    F12 = u'\x1B[24~',
    END = u'\x1B[1F',
    HOME = u'\x1B[1H',
    UP = u'\x1B[1A',
    DOWN = u'\x1B[1B',
    RIGHT = u'\x1B[1C',
    LEFT = u'\x1B[1D',
    INSERT = u'\x1B[2~',
    DELETE = u'\x1B[3~',
    PAGEUP = u'\x1B[5~',
    PAGEDOWN = u'\x1B[6~',
)

# CSI-based key codes
CSI_KEYS = dict(
    END = u'\x1B[F',
    HOME = u'\x1B[H',
    UP = u'\x1B[A',
    DOWN = u'\x1B[B',
    RIGHT = u'\x1B[C',
    LEFT = u'\x1B[D',
)

# SS3-based key codes
SS3_KEYS = dict(
    F1 = u'\x1BOP',
    F2 = u'\x1BOQ',
    F3 = u'\x1BOR',
    F4 = u'\x1BOS',
    END = u'\x1BOF',
    HOME = u'\x1BOH',
)

def _mod_csi(number):
    """Generate dict of modified CSI key sequences."""
    return {
        key: sequence[:-1] + u';%d' % (number,) + sequence[-1]
        for key, sequence in BASE_KEYS.items()
    }

# modified key codes
MOD_KEYS = {
    ('SHIFT',): _mod_csi(2),
    ('ALT',): _mod_csi(3),
    ('SHIFT', 'ALT'): _mod_csi(4),
    ('CTRL',): _mod_csi(5),
    ('SHIFT', 'CTRL'): _mod_csi(6),
    ('CTRL', 'ALT'): _mod_csi(7),
    ('SHIFT', 'CTRL', 'ALT'): _mod_csi(8),
}

# construct ansi to output mapping
ANSI_TO_KEYMOD = {
    sequence: (key, set(mods))
    for mods, mod_key_dict in MOD_KEYS.items()
    for key, sequence in mod_key_dict.items()
}
ANSI_TO_KEYMOD.update({sequence: (key, set()) for key, sequence in BASE_KEYS.items()})
ANSI_TO_KEYMOD.update({sequence: (key, set()) for key, sequence in CSI_KEYS.items()})
ANSI_TO_KEYMOD.update({sequence: (key, set()) for key, sequence in SS3_KEYS.items()})

# esc + char means alt+key; lowercase
ANSI_TO_KEYMOD.update({u'\x1b%c' % (c + 32,): (chr(c + 32), {'ALT'}) for c in range(65, 91)})
# uppercase
ANSI_TO_KEYMOD.update({u'\x1b%c' % (c,): (chr(c + 32), {'ALT', 'SHIFT'}) for c in range(65, 91)})
# digits, controls & everything else
ANSI_TO_KEYMOD.update({u'\x1b%c' % (c,): (chr(c), {'ALT'}) for c in range(0, 65)})
ANSI_TO_KEYMOD.update({u'\x1b%c' % (c,): (chr(c), {'ALT'}) for c in range(91, 128)})


# mapping of the first 8 attributes of the default CGA palette
# so that non-RGB terminals use sensible colours
# black, blue, green, cyan, red, magenta, yellow, white
EGA_TO_ANSI = (0, 4, 2, 6, 1, 5, 3, 7)

# default palette - these are in fact the 16 CGA colours
# this gets overwritten anyway
DEFAULT_PALETTE = (
    (0x00, 0x00, 0x00), (0x00, 0x00, 0xaa), (0x00, 0xaa, 0x00), (0x00, 0xaa, 0xaa),
    (0xaa, 0x00, 0x00), (0xaa, 0x00, 0xaa), (0xaa, 0x55, 0x00), (0xaa, 0xaa, 0xaa),
    (0x55, 0x55, 0x55), (0x55, 0x55, 0xff), (0x55, 0xff, 0x55), (0x55, 0xff, 0xff),
    (0xff, 0x55, 0x55), (0xff, 0x55, 0xff), (0xff, 0xff, 0x55), (0xff, 0xff, 0xff)
)


# output buffer for ioctl call
_sock_size = array.array('i', [0])


# standard unicode streams
if PY2:
    stdin = wrap_input_stream(sys.stdin)
    stdout = wrap_output_stream(sys.stdout)
    stderr = wrap_output_stream(sys.stderr)
else:
    stdin, stdout, stderr = sys.stdin, sys.stdout, sys.stderr


class PosixConsole(object):
    """POSIX-based console implementation."""

    def __init__(self):
        """Set up the console."""
        # buffer to save termios state
        if not sys.stdin.isatty() or not sys.stdout.isatty():
            raise EnvironmentError('Not a terminal')
        self._term_attr = termios.tcgetattr(sys.stdin.fileno())
        # preserve original terminal size
        self._orig_size = self.get_size()
        # input buffer
        self._read_buffer = deque()
        # palette
        self._palette = list(DEFAULT_PALETTE)

    ##########################################################################
    # terminal modes

    def get_size(self):
        """Get terminal size."""
        try:
            return struct.unpack(
                'HHHH', fcntl.ioctl(sys.stdout.fileno(), termios.TIOCGWINSZ, b'\0'*8)
            )[:2]
        except Exception:
            return 25, 80

    def set_raw(self):
        """Enter raw terminal mode (no echo, don't exit on ctrl-C, ...)."""
        tty.setraw(sys.stdin.fileno())

    def unset_raw(self):
        """Leave raw terminal mode."""
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSADRAIN, self._term_attr)


    ##########################################################################
    # ansi output

    def write(self, unicode_str):
        """Write unicode to console."""
        stdout.write(unicode_str)
        stdout.flush()

    def _emit_ansi(self, ansistr):
        """Emit escape code."""
        stdout.write(ansistr)
        stdout.flush()

    def set_caption(self, caption):
        """Set terminal caption."""
        self._emit_ansi(ANSI.SET_TITLE % (caption,))

    def resize(self, height, width):
        """Resize terminal."""
        self._emit_ansi(ANSI.RESIZE_TERM % (height, width))
        # start below the current output
        self.clear()

    def clear(self):
        """Clear the screen."""
        self._emit_ansi(
            ANSI.CLEAR_SCREEN +
            ANSI.MOVE_CURSOR % (1, 1)
        )

    def clear_row(self):
        """Clear the current row."""
        self._emit_ansi(ANSI.CLEAR_LINE)

    def show_cursor(self, block=False):
        """Show the cursor."""
        self._emit_ansi(
            ANSI.SHOW_CURSOR +
            ANSI.SET_CURSOR_SHAPE % (1 if block else 3,)
        )

    def hide_cursor(self):
        """Hide the cursor."""
        self._emit_ansi(ANSI.HIDE_CURSOR)

    def move_cursor_left(self, n):
        """Move cursor n cells to the left."""
        self._emit_ansi(ANSI.MOVE_N_LEFT % (n,))

    def move_cursor_right(self, n):
        """Move cursor n cells to the right."""
        self._emit_ansi(ANSI.MOVE_N_RIGHT % (n,))

    def move_cursor_to(self, row, col):
        """Move cursor to a new position."""
        self._emit_ansi(ANSI.MOVE_CURSOR % (row, col))

    def scroll_up(self, top, bottom):
        """Scroll the region between top and bottom one row up."""
        if bottom > top:
            self._emit_ansi(
                ANSI.SET_SCROLL_REGION % (top, bottom) +
                ANSI.SCROLL_UP % (1,) +
                ANSI.SET_SCROLL_SCREEN
            )

    def scroll_down(self, top, bottom):
        """Scroll the region between top and bottom one row down."""
        if bottom > top:
            self._emit_ansi(
                ANSI.SET_SCROLL_REGION % (top, bottom) +
                ANSI.SCROLL_DOWN % (1,) +
                ANSI.SET_SCROLL_SCREEN
            )

    def set_cursor_colour(self, colour):
        """Set the current cursor colour attribute."""
        try:
            rgb = self._palette[colour]
            self._emit_ansi(ANSI.SET_CURSOR_COLOUR % rgb)
        except KeyError:
            pass

    def reset(self):
        """Reset to defaults."""
        self._emit_ansi(
            ANSI.RESIZE_TERM % self._orig_size +
            u''.join(ANSI.RESET_PALETTE_ENTRY % (attr,) for attr in range(16)) +
            ANSI.SET_COLOUR % (0,) +
            ANSI.SHOW_CURSOR +
            ANSI.SET_CURSOR_COLOUR % (0xff, 0xff, 0xff) +
            ANSI.SET_CURSOR_SHAPE % (1,)
        )

    def set_attributes(self, fore, back, blink, underline):
        """Set current attributes."""
        # use "bold" ANSI colours for the upper 8 EGA attributes
        style = 90 if (fore > 8) else 30
        self._emit_ansi(
            ANSI.SET_COLOUR % (0,) +
            ANSI.SET_COLOUR % (40 + EGA_TO_ANSI[back],) +
            ANSI.SET_COLOUR % (style + EGA_TO_ANSI[fore % 8],)
        )
        if blink:
            self._emit_ansi(ANSI.SET_COLOUR % (5,))
        if underline:
            self._emit_ansi(ANSI.SET_COLOUR % (4,))

    def set_palette_entry(self, attr, red, green, blue):
        """Set palette entry for attribute (0--16)."""
        # keep a record, mainly for cursor colours
        self._palette[attr] = red, green, blue
        # set the ANSI palette
        self._emit_ansi(ANSI.SET_PALETTE_ENTRY % (
            8*(attr//8) + EGA_TO_ANSI[attr%8], red, green, blue
        ))

    ##########################################################################
    # input

    def key_pressed(self):
        """Return whether a character is ready to be read from the keyboard."""
        return select.select([sys.stdin], [], [], 0)[0] != []

    def read_key(self):
        """
        Read keypress from console. Non-blocking.
        Returns tuple (unicode, keycode, set of mods)
        """
        sequence = read_all_available(stdin)
        if sequence is None:
            # stream closed, send ctrl-d
            return u'\x04', 'd', {'CTRL'}
        elif not sequence:
            return u'', None, set()
        # ansi sequences start with ESC (\x1b), but let ESC by itself through
        if len(sequence) > 1 and sequence[0] == u'\x1b':
            # drop unrecognised sequences
            key, mod = ANSI_TO_KEYMOD.get(sequence, (u'', ()))
            return u'', key, mod
        else:
            return sequence, None, set()


def _is_console_app():
    """To see if we are a console app, check if we can treat stdin like a tty, file or socket."""
    if not sys.stdin.isatty():
        try:
            fcntl.ioctl(sys.stdin, termios.FIONREAD, _sock_size)
        except EnvironmentError:
            # maybe /dev/null, but not a real file or console
            return False
    return True

IS_CONSOLE_APP = _is_console_app()

try:
    console = PosixConsole()
except EnvironmentError:
    console = None

# don't crash into raw terminal
atexit.register(lambda: console.unset_raw() if console else None)


def read_all_available(stream):
    """Read all available characters from a stream; nonblocking; None if closed."""
    # this function works for everything on unix, and sockets on Windows
    instr = []
    # we're getting bytes counts for unicode which is pretty useless - so back to bytes
    try:
        encoding = stream.encoding
        stream = stream.buffer
    except:
        encoding = None
    # if buffer has characters/lines to read
    if select.select([stream], [], [], 0)[0]:
        # find number of bytes available
        fcntl.ioctl(stream, termios.FIONREAD, _sock_size)
        count = _sock_size[0]
        # and read them all
        c = stream.read(count)
        if not c and not instr:
            # break out, we're closed
            return None
        instr.append(c)
    if encoding:
        return b''.join(instr).decode(encoding, 'replace')
    return b''.join(instr)



"""
PC-BASIC - textscreen.py
Text operations

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import logging

from ...compat import iterchar, iteritems, int2byte

from ..base import signals
from ..base import error
from ..base import tokens as tk
from .. import values
from . import font
from .text import TextBuffer, TextRow
from .textbase import BottomBar, Cursor, ScrollArea


class TextScreen(object):
    """Text screen."""

    def __init__(self, queues, values, mode, capabilities, fonts, codepage, io_streams, sound):
        """Initialise text-related members."""
        self.queues = queues
        self._values = values
        self.codepage = codepage
        self.capabilities = capabilities
        # output redirection
        self._io_streams = io_streams
        # sound output needed for printing \a
        self.sound = sound
        # cursor
        self.cursor = Cursor(queues, mode, capabilities)
        # current row and column
        # overflow: true if we're on 80 but should be on 81
        self.current_row, self.current_col, self.overflow = 1, 1, False
        # text viewport parameters
        self.scroll_area = ScrollArea(mode)
        # writing on bottom row is allowed
        self._bottom_row_allowed = False
        # prepare fonts
        if not fonts:
            fonts = {8: {}}
        self.fonts = {
            height: font.Font(height, font_dict)
            for height, font_dict in iteritems(fonts)
        }
        # function key macros
        self.bottom_bar = BottomBar()

    def init_mode(self, mode, pixels, attr, vpagenum, apagenum):
        """Reset the text screen for new video mode."""
        self.mode = mode
        self.attr = attr
        self.apagenum = apagenum
        self.vpagenum = vpagenum
        # set up glyph cache and preload halfwidth glyphs (i.e. single-byte code points)
        self._glyphs = font.GlyphCache(self.mode, self.fonts, self.codepage, self.queues)
        # build the screen buffer
        self.text = TextBuffer(
            self.attr, self.mode.width, self.mode.height, self.mode.num_pages,
            self.codepage, do_fullwidth=(self.mode.font_height >= 14)
        )
        # pixel buffer
        self.pixels = pixels
        # redraw key line
        self.bottom_bar.redraw(self)
        # initialise text viewport & move cursor home
        self.scroll_area.init_mode(self.mode)
        self.set_pos(self.scroll_area.top, 1)
        # rebuild the cursor
        self.cursor.init_mode(self.mode, self.attr)

    def set_page(self, vpagenum, apagenum):
        """Set visible and active page."""
        self.vpagenum = vpagenum
        self.apagenum = apagenum

    def set_attr(self, attr):
        """Set attribute."""
        self.attr = attr

    def check_font_available(self, mode):
        """Raise IFC if no suitable font available for this mode."""
        if mode.font_height not in self.fonts:
            logging.warning(
                'No %d-pixel font available. Could not enter video mode %s.',
                mode.font_height, mode.name
            )
            raise error.BASICError(error.IFC)

    def rebuild(self):
        """Completely resubmit the text screen to the interface."""
        # send the glyph dict to interface if necessary
        self._glyphs.submit()
        # fix the cursor
        self.queues.video.put(signals.Event(
            signals.VIDEO_SET_CURSOR_SHAPE,
            (self.cursor.width, self.mode.font_height, self.cursor.from_line, self.cursor.to_line)
        ))
        self.queues.video.put(signals.Event(
            signals.VIDEO_MOVE_CURSOR, (self.current_row, self.current_col)
        ))
        if self.mode.is_text_mode:
            attr = self.text.get_attr(self.apagenum, self.current_row, self.current_col)
            fore, _, _, _ = self.mode.split_attr(attr & 0xf)
        else:
            fore, _, _, _ = self.mode.split_attr(self.mode.cursor_index or self.attr)
        self.queues.video.put(signals.Event(signals.VIDEO_SET_CURSOR_ATTR, (fore,)))
        self.cursor.reset_visibility()
        # redraw the text screen and rebuild text buffers in video plugin
        for pagenum in range(self.mode.num_pages):
            for row in range(self.mode.height):
                self.refresh_range(pagenum, row+1, 1, self.mode.width, text_only=True)
            # redraw graphics
            if not self.mode.is_text_mode:
                self.queues.video.put(signals.Event(
                    signals.VIDEO_PUT_RECT, (
                        pagenum, 0, 0, self.mode.pixel_width-1, self.mode.pixel_height-1,
                        self.pixels.pages[pagenum].buffer
                    )
                ))

    def __repr__(self):
        """Return an ascii representation of the screen buffer (for debugging)."""
        return repr(self.text)

    ##########################################################################

    def write(self, s, scroll_ok=True, do_echo=True):
        """Write a string to the screen at the current position."""
        if do_echo:
            # CR -> CRLF, CRLF -> CRLF LF
            self._io_streams.write(b''.join([(b'\r\n' if c == b'\r' else c) for c in iterchar(s)]))
        last = b''
        # if our line wrapped at the end before, it doesn't anymore
        self.text.pages[self.apagenum].row[self.current_row-1].wrap = False
        for c in iterchar(s):
            row, col = self.current_row, self.current_col
            if c == b'\t':
                # TAB
                num = (8 - (col - 1 - 8 * int((col-1) // 8)))
                for _ in range(num):
                    self.write_char(b' ')
            elif c == b'\n':
                # LF
                # exclude CR/LF
                if last != b'\r':
                    # LF connects lines like word wrap
                    self.text.pages[self.apagenum].row[row-1].wrap = True
                    self.set_pos(row + 1, 1, scroll_ok)
            elif c == b'\r':
                # CR
                self.text.pages[self.apagenum].row[row-1].wrap = False
                self.set_pos(row + 1, 1, scroll_ok)
            elif c == b'\a':
                # BEL
                self.sound.beep()
            elif c == b'\x0B':
                # HOME
                self.set_pos(1, 1, scroll_ok)
            elif c == b'\x0C':
                # CLS
                self.clear_view()
            elif c == b'\x1C':
                # RIGHT
                self.set_pos(row, col + 1, scroll_ok)
            elif c == b'\x1D':
                # LEFT
                self.set_pos(row, col - 1, scroll_ok)
            elif c == b'\x1E':
                # UP
                self.set_pos(row - 1, col, scroll_ok)
            elif c == b'\x1F':
                # DOWN
                self.set_pos(row + 1, col, scroll_ok)
            else:
                # includes \b, \0, and non-control chars
                self.write_char(c)
            last = c

    def write_line(self, s=b'', scroll_ok=True, do_echo=True):
        """Write a string to the screen and end with a newline."""
        self.write(b'%s\r' % (s,), scroll_ok, do_echo)

    def list_line(self, line, newline=True):
        """Print a line from a program listing or EDIT prompt."""
        # no wrap if 80-column line, clear row before printing.
        # replace LF CR with LF
        line = line.replace(b'\n\r', b'\n')
        cuts = line.split(b'\n')
        for i, l in enumerate(cuts):
            # clear_line looks back along wraps, use screen.clear_from instead
            self.clear_from(self.current_row, 1)
            self.write(l)
            if i != len(cuts) - 1:
                self.write(b'\n')
        if newline:
            self.write_line()
        # remove wrap after 80-column program line
        if len(line) == self.mode.width and self.current_row > 2:
            self.text.pages[self.apagenum].row[self.current_row-3].wrap = False

    def write_char(self, c, do_scroll_down=False):
        """Put one character at the current position."""
        # check if scroll& repositioning needed
        if self.overflow:
            self.current_col += 1
            self.overflow = False
        # see if we need to wrap and scroll down
        self._check_wrap(do_scroll_down)
        # move cursor and see if we need to scroll up
        self._check_pos(scroll_ok=True)
        # put the character
        self.put_char_attr(self.apagenum, self.current_row, self.current_col, c, self.attr)
        # adjust end of line marker
        if (self.current_col > self.text.pages[self.apagenum].row[self.current_row-1].end):
            self.text.pages[self.apagenum].row[self.current_row-1].end = self.current_col
        # move cursor. if on col 80, only move cursor to the next row
        # when the char is printed
        if self.current_col < self.mode.width:
            self.current_col += 1
        else:
            self.overflow = True
        # move cursor and see if we need to scroll up
        self._check_pos(scroll_ok=True)

    def _check_wrap(self, do_scroll_down):
        """Wrap if we need to."""
        if self.current_col > self.mode.width:
            if self.current_row < self.mode.height:
                # wrap line
                self.text.pages[self.apagenum].row[self.current_row-1].wrap = True
                if do_scroll_down:
                    # scroll down (make space by shifting the next rows down)
                    if self.current_row < self.scroll_area.bottom:
                        self.scroll_down(self.current_row+1)
                # move cursor and reset cursor attribute
                self._move_cursor(self.current_row + 1, 1)
            else:
                self.current_col = self.mode.width

    def start_line(self):
        """Move the cursor to the start of the next line, this line if empty."""
        if self.current_col != 1:
            self._io_streams.write(b'\r\n')
            self._check_pos(scroll_ok=True)
            self.set_pos(self.current_row + 1, 1)
        # ensure line above doesn't wrap
        self.text.pages[self.apagenum].row[self.current_row-2].wrap = False

    ###########################################################################
    # cursor position

    def set_pos(self, to_row, to_col, scroll_ok=True):
        """Set the current position."""
        self.overflow = False
        self.current_row, self.current_col = to_row, to_col
        # move cursor and reset cursor attribute
        # this may alter self.current_row, self.current_col
        self._check_pos(scroll_ok)

    def incr_pos(self):
        """Increase the current position by a char width."""
        # on a trail byte: just go one to the right
        width = self.text.get_charwidth(self.apagenum, self.current_row, self.current_col) or 1
        self.set_pos(self.current_row, self.current_col + width, scroll_ok=False)

    def decr_pos(self):
        """Decrease the current position by a char width."""
        # previous is trail byte: go two to the left
        # lead byte: go three to the left
        width = self.text.get_charwidth(self.apagenum, self.current_row, self.current_col-1)
        if width == 0:
            skip = 2
        elif width == 2:
            skip = 3
        else:
            skip = 1
        self.set_pos(self.current_row, self.current_col - skip, scroll_ok=False)

    def _check_pos(self, scroll_ok=True):
        """Check if we have crossed the screen boundaries and move as needed."""
        oldrow, oldcol = self.current_row, self.current_col
        if self._bottom_row_allowed:
            if self.current_row == self.mode.height:
                self.current_col = min(self.mode.width, self.current_col)
                if self.current_col < 1:
                    self.current_col += 1
                self._move_cursor(self.current_row, self.current_col)
                return self.current_col == oldcol
            else:
                # if row > height, we also end up here
                # (eg if we do INPUT on the bottom row)
                # adjust viewport if necessary
                self._bottom_row_allowed = False
        # see if we need to move to the next row
        if self.current_col > self.mode.width:
            if self.current_row < self.scroll_area.bottom or scroll_ok:
                # either we don't need to scroll, or we're allowed to
                self.current_col -= self.mode.width
                self.current_row += 1
            else:
                # we can't scroll, so we just stop at the right border
                self.current_col = self.mode.width
        # see if we need to move a row up
        elif self.current_col < 1:
            if self.current_row > self.scroll_area.top:
                self.current_col += self.mode.width
                self.current_row -= 1
            else:
                self.current_col = 1
        # see if we need to scroll
        if self.current_row > self.scroll_area.bottom:
            if scroll_ok:
                self.scroll()
            self.current_row = self.scroll_area.bottom
        elif self.current_row < self.scroll_area.top:
            self.current_row = self.scroll_area.top
        self._move_cursor(self.current_row, self.current_col)
        # signal position change
        return (self.current_row == oldrow and self.current_col == oldcol)

    def _move_cursor(self, row, col):
        """Move the cursor to a new position."""
        self.current_row, self.current_col = row, col
        # set halfwidth/fullwidth cursor
        width = self.text.get_charwidth(self.apagenum, self.current_row, self.current_col)
        self.cursor.set_width(width)
        # set the cursor's attribute to that of the current location
        fore, _, _, _ = self.mode.split_attr(
            0xf & self.text.get_attr(self.apagenum, self.current_row, self.current_col)
        )
        self.cursor.reset_attr(fore)
        self.queues.video.put(signals.Event(
            signals.VIDEO_MOVE_CURSOR, (self.current_row, self.current_col))
        )

    def move_to_end(self):
        """Jump to end of logical line; follow wraps (END)."""
        row = self.text.find_end_of_line(self.apagenum, self.current_row)
        if self.text.pages[self.apagenum].row[row-1].end == self.mode.width:
            self.set_pos(row, self.text.pages[self.apagenum].row[row-1].end)
            self.overflow = True
        else:
            self.set_pos(row, self.text.pages[self.apagenum].row[row-1].end+1)

    ###########################################################################

    def put_char_attr(self, pagenum, row, col, c, attr, one_only=False):
        """Put a byte to the screen, redrawing as necessary."""
        if not self.mode.is_text_mode:
            attr = attr & 0xf
        start, stop = self.text.put_char_attr(pagenum, row, col, c, attr)
        if one_only:
            stop = start
        # update the screen
        self.refresh_range(pagenum, row, start, stop)

    ###########################################################################

    def refresh_range(self, pagenum, row, start, stop, text_only=False):
        """Redraw a section of a screen row, assuming DBCS buffer has been set."""
        therow = self.text.pages[pagenum].row[row-1]
        col = start
        while col <= stop:
            r, c = row, col
            char, attr = self.text.get_fullchar_attr(pagenum, row, col)
            col += len(char)
            # ensure glyph is stored
            self._glyphs.check_char(char)
            fore, back, blink, underline = self.mode.split_attr(attr)
            self.queues.video.put(signals.Event(
                signals.VIDEO_PUT_GLYPH, (
                    pagenum, r, c, self.codepage.to_unicode(char, u'\0'),
                    len(char) > 1, fore, back, blink, underline,
                )
            ))
            if not self.mode.is_text_mode and not text_only:
                # update pixel buffer
                x0, y0, x1, y1, sprite = self._glyphs.get_sprite(r, c, char, fore, back)
                self.pixels.pages[self.apagenum].put_rect(x0, y0, x1, y1, sprite, tk.PSET)
                self.queues.video.put(signals.Event(
                    signals.VIDEO_PUT_RECT, (self.apagenum, x0, y0, x1, y1, sprite)
                ))

    def _redraw_row(self, start, row, wrap=True):
        """Draw the screen row, wrapping around and reconstructing DBCS buffer."""
        while True:
            for i in range(start, self.text.pages[self.apagenum].row[row-1].end):
                # redrawing changes colour attributes to current foreground (cf. GW)
                # don't update all dbcs chars behind at each put
                char = int2byte(self.text.get_char(self.apagenum, row, i+1))
                self.put_char_attr(self.apagenum, row, i+1, char, self.attr, one_only=True)
            if (
                    wrap and self.text.pages[self.apagenum].row[row-1].wrap and
                    row >= 0 and row < self.text.height-1
                ):
                row += 1
                start = 0
            else:
                break

    def clear_from(self, srow, scol):
        """Clear from given position to end of logical line (CTRL+END)."""
        self.text.pages[self.apagenum].row[srow-1].clear_from(scol, self.attr)
        row = srow
        # can use self.text.find_end_of_line
        while self.text.pages[self.apagenum].row[row-1].wrap:
            row += 1
            self.text.pages[self.apagenum].row[row-1].clear(self.attr)
        for r in range(row, srow, -1):
            self.text.pages[self.apagenum].row[r-1].wrap = False
            self.scroll(r)
        therow = self.text.pages[self.apagenum].row[srow-1]
        therow.wrap = False
        self.set_pos(srow, scol)
        save_end = therow.end
        therow.end = self.mode.width
        if scol > 1:
            self._redraw_row(scol-1, srow)
        else:
            # inelegant: we're clearing the text buffer for a second time now
            self.clear_rows(srow, srow)
        therow.end = save_end

    ###########################################################################
    # clearing text screen

    def clear_rows(self, start, stop):
        """Clear text and graphics on given (inclusive) text row range."""
        for r in self.text.pages[self.apagenum].row[start-1:stop]:
            r.clear(self.attr)
            # can't we just do this in row.clear?
            r.wrap = False
        if not self.mode.is_text_mode:
            x0, y0, x1, y1 = self.mode.text_to_pixel_area(start, 1, stop, self.mode.width)
            # background attribute must be 0 in graphics mode
            self.pixels.pages[self.apagenum].fill_rect(x0, y0, x1, y1, 0)
        _, back, _, _ = self.mode.split_attr(self.attr)
        self.queues.video.put(signals.Event(signals.VIDEO_CLEAR_ROWS, (back, start, stop)))

    def _clear_area(self, start_row, stop_row):
        """Clear the screen or the scroll area."""
        if self.capabilities in ('vga', 'ega', 'cga', 'cga_old'):
            # keep background, set foreground to 7
            attr_save = self.attr
            self.set_attr(attr_save & 0x70 | 0x7)
        self.clear_rows(start_row, stop_row)
        # ensure the cursor is shown in the right position
        self.set_pos(start_row, 1)
        if self.capabilities in ('vga', 'ega', 'cga', 'cga_old'):
            # restore attr
            self.set_attr(attr_save)

    def clear_view(self):
        """Clear the scroll area."""
        self._clear_area(self.scroll_area.top, self.scroll_area.bottom)

    def clear(self):
        """Clear the screen."""
        self._clear_area(1, self.mode.height)

    ###########################################################################
    # text viewport / scroll area

    def _set_scroll_area(self, start, stop):
        """Set the scroll area."""
        self.scroll_area.set(start, stop)
        #set_pos(start, 1)
        self.overflow = False
        self._move_cursor(start, 1)

    ###########################################################################
    # scrolling

    def scroll(self, from_line=None):
        """Scroll the scroll region up by one line, starting at from_line."""
        if from_line is None:
            from_line = self.scroll_area.top
        _, back, _, _ = self.mode.split_attr(self.attr)
        self.queues.video.put(signals.Event(
            signals.VIDEO_SCROLL_UP, (from_line, self.scroll_area.bottom, back)
        ))
        if self.current_row > from_line:
            self._move_cursor(self.current_row - 1, self.current_col)
        # sync buffers with the new screen reality:
        self.text.scroll_up(self.apagenum, from_line, self.scroll_area.bottom, self.attr)
        if not self.mode.is_text_mode:
            sx0, sy0, sx1, sy1 = self.mode.text_to_pixel_area(from_line+1, 1,
                self.scroll_area.bottom, self.mode.width)
            tx0, ty0, _, _ = self.mode.text_to_pixel_area(from_line, 1,
                self.scroll_area.bottom-1, self.mode.width)
            self.pixels.pages[self.apagenum].move_rect(sx0, sy0, sx1, sy1, tx0, ty0)

    def scroll_down(self, from_line):
        """Scroll the scroll region down by one line, starting at from_line."""
        _, back, _, _ = self.mode.split_attr(self.attr)
        self.queues.video.put(signals.Event(
            signals.VIDEO_SCROLL_DOWN, (from_line, self.scroll_area.bottom, back)
        ))
        if self.current_row >= from_line:
            self._move_cursor(self.current_row + 1, self.current_col)
        # sync buffers with the new screen reality:
        self.text.scroll_down(self.apagenum, from_line, self.scroll_area.bottom, self.attr)
        if not self.mode.is_text_mode:
            sx0, sy0, sx1, sy1 = self.mode.text_to_pixel_area(from_line, 1,
                self.scroll_area.bottom-1, self.mode.width)
            tx0, ty0, _, _ = self.mode.text_to_pixel_area(from_line+1, 1,
                self.scroll_area.bottom, self.mode.width)
            self.pixels.pages[self.apagenum].move_rect(sx0, sy0, sx1, sy1, tx0, ty0)

    ###########################################################################
    # console operations

    def delete_fullchar(self):
        """Delete the character (single/double width) at the current position."""
        row, col = self.current_row, self.current_col
        cwidth = self.text.get_charwidth(self.apagenum, self.current_row, self.current_col)
        if cwidth == 0:
            logging.debug('DBCS trail byte delete at %d, %d.', self.current_row, self.current_col)
            self.set_pos(self.current_row, self.current_col-1)
            cwidth = 2
        text = self.text.get_logical_line(self.apagenum, self.current_row, self.current_col)
        lastrow = self.text.find_end_of_line(self.apagenum, self.current_row)
        if self.current_col > self.text.pages[self.apagenum].row[self.current_row-1].end:
            # past the end. if this row ended with LF, attach next row and scroll further rows up
            # if not, do nothing
            if not self.text.pages[self.apagenum].row[self.current_row-1].wrap:
                return
            # else: LF case; scroll up without changing attributes
            self.text.pages[self.apagenum].row[self.current_row-1].wrap = (
                self.text.pages[self.apagenum].row[self.current_row].wrap
            )
            self.scroll(self.current_row + 1)
            # redraw from the LF
            self._rewrite_for_delete(text[1:] + b' ' * cwidth)
        else:
            # rewrite the contents (with the current attribute!)
            self._rewrite_for_delete(text[cwidth:] + b' ' * cwidth)
            # if last row was empty, scroll up.
            if (
                    self.text.pages[self.apagenum].row[lastrow-1].end == 0 and
                    self.text.pages[self.apagenum].row[lastrow-2].wrap
                ):
                self.text.pages[self.apagenum].row[lastrow-2].wrap = False
                self.scroll(lastrow)
        # restore original position
        self.set_pos(row, col)

    def _rewrite_for_delete(self, text):
        """Rewrite text contents (with the current attribute)."""
        for c in iterchar(text):
            if c == b'\n':
                self.put_char_attr(
                    self.apagenum, self.current_row, self.current_col, b' ', self.attr
                )
                self.text.pages[self.apagenum].row[self.current_row-1].end = self.current_col-1
                break
            else:
                self.put_char_attr(self.apagenum, self.current_row, self.current_col, c, self.attr)
                self.set_pos(self.current_row, self.current_col+1)
        else:
            # we're on the position after the additional space
            if self.current_col == 1:
                self.current_row -= 1
                self.text.pages[self.apagenum].row[self.current_row-1].end = self.mode.width - 1
            else:
                # adjust row end
                self.text.pages[self.apagenum].row[self.current_row-1].end = self.current_col - 2

    def insert_fullchars(self, sequence):
        """Insert one or more single- or double-width characters and adjust cursor."""
        # insert one at a time at cursor location
        # to let cursor position logic deal with scrolling
        for c in iterchar(sequence):
            if self._insert_fullchar_at(self.current_row, self.current_col, c, self.attr):
                # move cursor by one character
                # this will move to next row when necessary
                self.set_pos(self.current_row, self.current_col+1)

    def _insert_fullchar_at(self, row, col, c, attr):
        """Insert one single- or double-width character at the given position."""
        therow = self.text.pages[self.apagenum].row[row-1]
        if therow.end < self.mode.width:
            therow.buf.insert(col-1, (c, attr))
            c, attr = therow.buf.pop()
            if therow.end > col-1:
                therow.end += 1
            else:
                therow.end = col
            self._redraw_row(col-1, row)
            if  therow.wrap and therow.end == self.mode.width:
                self.scroll_down(row+1)
                therow.wrap = True
            return True
        else:
            # pushing the end of the row past the screen edge
            # if we're not a wrapping line, make space by scrolling
            if not therow.wrap and row < self.scroll_area.bottom:
                self.scroll_down(row+1)
                therow.wrap = True
            if row >= self.scroll_area.bottom:
                # once the end of the line hits the bottom, start scrolling the start of the line up
                # until that hist the top of the screen. After that, stop inserting & drop chars
                start = self.text.find_start_of_line(self.apagenum, self.current_row)
                if start > self.scroll_area.top:
                    self.scroll()
                    row -= 1
                else:
                    return False
            therow.buf.insert(col-1, (c, attr))
            c, attr = therow.buf.pop()
            self._redraw_row(col-1, row)
            # insert the character in the next row
            return self._insert_fullchar_at(row+1, 1, c, attr)

    def line_feed(self):
        """Move the remainder of the line to the next row and wrap (LF)."""
        if self.current_col < self.text.pages[self.apagenum].row[self.current_row-1].end:
            # insert characters, preserving cursor position
            cursor = self.current_row, self.current_col
            self.insert_fullchars(b' ' * (self.mode.width-self.current_col+1))
            self.set_pos(*cursor, scroll_ok=False)
            # adjust end of line and wrapping flag - LF connects lines like word wrap
            self.text.pages[self.apagenum].row[self.current_row-1].end = self.current_col - 1
            self.text.pages[self.apagenum].row[self.current_row-1].wrap = True
            # cursor stays in place after line feed!
        else:
            # find last row in logical line
            end = self.text.find_end_of_line(self.apagenum, self.current_row)
            # if the logical line hits the bottom, start scrolling up to make space...
            if end >= self.scroll_area.bottom:
                # ... until the it also hits the top; then do nothing
                start = self.text.find_start_of_line(self.apagenum, self.current_row)
                if start > self.scroll_area.top:
                    self.scroll()
                else:
                    return
            # self.current_row has changed, don't use row var
            if self.current_row < self.mode.height:
                self.scroll_down(self.current_row+1)
                # if we were already a wrapping row, make sure the new empty row wraps
                #if self.text.pages[self.apagenum].row[self.current_row-1].wrap:
                #    self.text.pages[self.apagenum].row[self.current_row].wrap = True
            # ensure the current row now wraps
            self.text.pages[self.apagenum].row[self.current_row-1].wrap = True
            # cursor moves to start of next line
            self.set_pos(self.current_row+1, 1)

    ###########################################################################
    # vpage text retrieval

    def print_screen(self, target_file):
        """Output the visible page to file in raw bytes."""
        if not target_file:
            return
        for line in self.text.get_text_raw(self.vpagenum):
            target_file.write_line(line.replace(b'\0', b' '))

    def copy_clipboard(self, start_row, start_col, stop_row, stop_col, is_mouse_selection):
        """Copy selected screen area to clipboard."""
        clips = self.text.get_text_logical(
            self.vpagenum, start_row, start_col, stop_row, stop_col
        )
        text = u'\n'.join(self.codepage.str_to_unicode(clip) for clip in clips)
        self.queues.video.put(signals.Event(
                signals.VIDEO_SET_CLIPBOARD_TEXT, (text, is_mouse_selection)
        ))

    ###########################################################################
    # text screen callbacks

    def locate_(self, args):
        """LOCATE: Set cursor position, shape and visibility."""
        args = list(None if arg is None else values.to_int(arg) for arg in args)
        args = args + [None] * (5-len(args))
        row, col, cursor, start, stop = args
        row = self.current_row if row is None else row
        col = self.current_col if col is None else col
        cmode = self.mode
        error.throw_if(row == cmode.height and self.bottom_bar.visible)
        if self.scroll_area.active:
            error.range_check(self.scroll_area.top, self.scroll_area.bottom, row)
        else:
            error.range_check(1, cmode.height, row)
        error.range_check(1, cmode.width, col)
        if row == cmode.height:
            # temporarily allow writing on last row
            self._bottom_row_allowed = True
        self.set_pos(row, col, scroll_ok=False)
        if cursor is not None:
            error.range_check(0, (255 if self.capabilities in ('pcjr', 'tandy') else 1), cursor)
            # set cursor visibility - this should set the flag but have no effect in graphics modes
            self.cursor.set_visibility(cursor != 0)
        error.throw_if(start is None and stop is not None)
        if stop is None:
            stop = start
        if start is not None:
            error.range_check(0, 31, start, stop)
            # cursor shape only has an effect in text mode
            if cmode.is_text_mode:
                self.cursor.set_shape(start, stop)

    def csrlin_(self, args):
        """CSRLIN: get the current screen row."""
        list(args)
        if (self.overflow and self.current_col == self.mode.width and
                                    self.current_row < self.scroll_area.bottom):
            # in overflow position, return row+1 except on the last row
            csrlin = self.current_row + 1
        else:
            csrlin = self.current_row
        return self._values.new_integer().from_int(csrlin)

    def pos_(self, args):
        """POS: get the current screen column."""
        list(args)
        if self.current_col == self.mode.width and self.overflow:
            # in overflow position, return column 1.
            pos = 1
        else:
            pos = self.current_col
        return self._values.new_integer().from_int(pos)

    def screen_fn_(self, args):
        """SCREEN: get char or attribute at a location."""
        row = values.to_integer(next(args))
        col = values.to_integer(next(args))
        want_attr = next(args)
        if want_attr is not None:
            want_attr = values.to_integer(want_attr)
            want_attr = want_attr.to_int()
            error.range_check(0, 255, want_attr)
        row, col = row.to_int(), col.to_int()
        error.range_check(0, self.mode.height, row)
        error.range_check(0, self.mode.width, col)
        error.throw_if(row == 0 and col == 0)
        list(args)
        row = row or 1
        col = col or 1
        if self.scroll_area.active:
            error.range_check(self.scroll_area.top, self.scroll_area.bottom, row)
        if want_attr:
            if not self.mode.is_text_mode:
                result = 0
            else:
                result = self.text.get_attr(self.apagenum, row, col)
        else:
            result = self.text.get_char(self.apagenum, row, col)
        return self._values.new_integer().from_int(result)

    def view_print_(self, args):
        """VIEW PRINT: set scroll region."""
        start, stop = (None if arg is None else values.to_int(arg) for arg in args)
        if start is None and stop is None:
            self.scroll_area.unset()
        else:
            if self.capabilities in ('pcjr', 'tandy') and not self.bottom_bar.visible:
                max_line = 25
            else:
                max_line = 24
            error.range_check(1, max_line, start, stop)
            error.throw_if(stop < start)
            self._set_scroll_area(start, stop)



"""
PC-BASIC - operators.py
Numeric and string operators

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from ..base import tokens as tk
from .. import values


# operators and precedence
# key is tuple (token, nargs)
PRECEDENCE = {
    tk.O_CARET: 12,
    tk.O_TIMES: 11,
    tk.O_DIV: 11,
    tk.O_INTDIV: 10,
    tk.MOD: 9,
    tk.O_PLUS: 8,
    tk.O_MINUS: 8,
    tk.O_GT: 7,
    tk.O_EQ: 7,
    tk.O_LT: 7,
    tk.O_GT + tk.O_EQ: 7,
    tk.O_EQ + tk.O_GT: 7,
    tk.O_LT + tk.O_EQ: 7,
    tk.O_EQ + tk.O_LT: 7,
    tk.O_LT + tk.O_GT: 7,
    tk.O_GT + tk.O_LT: 7,
    tk.NOT: 6,
    tk.AND: 5,
    tk.OR: 4,
    tk.XOR: 3,
    tk.EQV: 2,
    tk.IMP: 1,
}
OPERATORS = set(PRECEDENCE)

# can be combined like <> >=
COMBINABLE = (tk.O_LT, tk.O_EQ, tk.O_GT)

# unary operators
UNARY = {
    tk.O_MINUS: values.neg,
    tk.O_PLUS: lambda x: x,
    tk.NOT: values.not_,
}

# binary operators
BINARY = {
    tk.O_CARET: values.pow,
    tk.O_TIMES: values.mul,
    tk.O_DIV: values.div,
    tk.O_INTDIV: values.intdiv,
    tk.MOD: values.mod_,
    tk.O_PLUS: values.add,
    tk.O_MINUS: values.sub,
    tk.O_GT: values.gt,
    tk.O_EQ: values.eq,
    tk.O_LT: values.lt,
    tk.O_GT + tk.O_EQ: values.gte,
    tk.O_EQ + tk.O_GT: values.gte,
    tk.O_LT + tk.O_EQ: values.lte,
    tk.O_EQ + tk.O_LT: values.lte,
    tk.O_LT + tk.O_GT: values.neq,
    tk.O_GT + tk.O_LT: values.neq,
    tk.AND: values.and_,
    tk.OR: values.or_,
    tk.XOR: values.xor_,
    tk.EQV: values.eqv_,
    tk.IMP: values.imp_,
}



"""
PC-BASIC - display.py
Display mode and colour palette operations

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import struct

from ..base import signals
from ..base import error
from .. import values
from . import graphics
from . import modes

from .textscreen import TextScreen
from .pixels import PixelBuffer
from .modes import Video


#######################################################################################
# palette

class Palette(object):
    """Colour palette."""

    def __init__(self, queues, mode, capabilities, memory):
        """Initialise palette."""
        self.capabilities = capabilities
        self._memory = memory
        self._queues = queues
        self.mode = mode
        self.set_all(mode.palette, check_mode=False)

    def init_mode(self, mode):
        """Initialise for new mode."""
        self.mode = mode
        self.set_all(mode.palette, check_mode=False)

    def set_entry(self, index, colour, check_mode=True):
        """Set a new colour for a given attribute."""
        mode = self.mode
        if check_mode and not self.mode_allows_palette(mode):
            return
        self.palette[index] = colour
        self.rgb_palette[index] = mode.colours[colour]
        if mode.colours1:
            self.rgb_palette1[index] = mode.colours1[colour]
        self._queues.video.put(signals.Event(
            signals.VIDEO_SET_PALETTE, (self.rgb_palette, self.rgb_palette1, None)
        ))

    def get_entry(self, index):
        """Retrieve the colour for a given attribute."""
        return self.palette[index]

    def set_all(self, new_palette, check_mode=True):
        """Set the colours for all attributes."""
        if check_mode and new_palette and not self.mode_allows_palette(self.mode):
            return
        self.palette = list(new_palette)
        self.rgb_palette = [self.mode.colours[i] for i in self.palette]
        if self.mode.colours1:
            self.rgb_palette1 = [self.mode.colours1[i] for i in self.palette]
        else:
            self.rgb_palette1 = None
        self._queues.video.put(signals.Event(
            signals.VIDEO_SET_PALETTE, (self.rgb_palette, self.rgb_palette1, None)
        ))

    def mode_allows_palette(self, mode):
        """Check if the video mode allows palette change."""
        # effective palette change is an error in CGA
        if self.capabilities in ('cga', 'cga_old', 'mda', 'hercules', 'olivetti'):
            raise error.BASICError(error.IFC)
        # ignore palette changes in Tandy/PCjr SCREEN 0
        elif self.capabilities in ('tandy', 'pcjr') and mode.is_text_mode:
            return False
        else:
            return True

    def palette_(self, args):
        """PALETTE: assign colour to attribute."""
        attrib = next(args)
        if attrib is not None:
            attrib = values.to_int(attrib)
        colour = next(args)
        if colour is not None:
            colour = values.to_int(colour)
        list(args)
        if attrib is None and colour is None:
            self.set_all(self.mode.palette)
        else:
            # can't set blinking colours separately
            num_palette_entries = self.mode.num_attr if self.mode.num_attr != 32 else 16
            error.range_check(0, num_palette_entries-1, attrib)
            colour = (colour+1) % 256 -1
            error.range_check(-1, len(self.mode.colours)-1, colour)
            if colour != -1:
                self.set_entry(attrib, colour)

    def palette_using_(self, args):
        """PALETTE USING: set palette from array buffer."""
        array_name, start_indices = next(args)
        array_name = self._memory.complete_name(array_name)
        list(args)
        num_palette_entries = self.mode.num_attr if self.mode.num_attr != 32 else 16
        try:
            dimensions = self._memory.arrays.dimensions(array_name)
        except KeyError:
            raise error.BASICError(error.IFC)
        error.throw_if(array_name[-1:] != values.INT, error.TYPE_MISMATCH)
        lst = self._memory.arrays.view_full_buffer(array_name)
        start = self._memory.arrays.index(start_indices, dimensions)
        error.throw_if(self._memory.arrays.array_len(dimensions) - start < num_palette_entries)
        new_palette = []
        for i in range(num_palette_entries):
            offset = (start+i) * 2
            ## signed int, as -1 means don't set
            val, = struct.unpack('<h', lst[offset:offset+2])
            error.range_check(-1, len(self.mode.colours)-1, val)
            new_palette.append(val if val > -1 else self.get_entry(i))
        self.set_all(new_palette)


#######################################################################################
# display

class Display(object):
    """Display and video mode manipulation operations."""

    def __init__(self, queues, values, input_methods, memory,
                initial_width, video_mem_size, capabilities, monitor, sound, io_streams,
                low_intensity, screen_aspect, codepage, fonts):
        """Initialise the display."""
        self.queues = queues
        self._values = values
        self._memory = memory
        # low level settings
        self.video = Video(
            capabilities, monitor, low_intensity, screen_aspect, video_mem_size
        )
        self.capabilities = self.video.capabilities
        # video mode settings
        self._mode_nr, self.colorswitch, self.apagenum, self.vpagenum = 0, 1, 0, 0
        # prepare video modes
        self.mode = self.video.get_mode(0, initial_width)
        # current attribute
        self.attr = 7
        # border attribute
        self.border_attr = 0
        # text screen
        self.text_screen = TextScreen(
            self.queues, self._values, self.mode, self.capabilities,
            fonts, codepage, io_streams, sound
        )
        # graphics operations
        self.drawing = graphics.Drawing(self.queues, input_methods, self._values, self._memory)
        # colour palette
        self.palette = Palette(self.queues, self.mode, self.capabilities, self._memory)
        # initialise a fresh textmode screen
        self._set_mode(self.mode, 0, 1, 0, 0)

    ###########################################################################
    # video modes

    def screen(self, new_mode, new_colorswitch, new_apagenum, new_vpagenum,
               erase=1, new_width=None, force_reset=False):
        """Change the video mode, colourburst, visible or active page."""
        # reset palette happens even if the SCREEN call fails
        self.palette.init_mode(self.mode)
        # set default arguments
        new_mode = self._mode_nr if (new_mode is None) else new_mode
        # set colorswitch
        if new_colorswitch is None:
            new_colorswitch = True
            if self.capabilities == 'pcjr':
                new_colorswitch = False
            elif self.capabilities == 'tandy':
                new_colorswitch = not new_mode
        new_colorswitch = bool(new_colorswitch)
        if new_mode == 0 and new_width is None:
            # if we switch out of a 20-col mode (Tandy screen 3), switch to 40-col.
            # otherwise, width persists on change to screen 0
            new_width = 40 if (self.mode.width == 20) else self.mode.width
        # retrieve the specs for the new video mode
        info = self.video.get_mode(new_mode, new_width)
        # vpage and apage nums are persistent on mode switch with SCREEN
        # on pcjr only, reset page to zero if current page number would be too high.
        # in other adapters, that's going to raise an IFC later on.
        if new_vpagenum is None:
            new_vpagenum = self.vpagenum
            if (self.capabilities == 'pcjr' and new_vpagenum >= info.num_pages):
                new_vpagenum = 0
        if new_apagenum is None:
            new_apagenum = self.apagenum
            if (self.capabilities == 'pcjr' and new_apagenum >= info.num_pages):
                new_apagenum = 0
        if (
                (not info.is_text_mode and info.name != self.mode.name) or
                (info.is_text_mode and not self.mode.is_text_mode) or
                (info.width != self.mode.width) or
                (new_colorswitch != self.colorswitch) or force_reset
            ):
            self._set_mode(
                    info, new_mode, new_colorswitch, new_apagenum, new_vpagenum, erase)
        else:
            # only switch pages
            if (new_apagenum >= info.num_pages or new_vpagenum >= info.num_pages):
                raise error.BASICError(error.IFC)
            self.set_page(new_vpagenum, new_apagenum)

    def _set_mode(self, spec, new_mode, new_colorswitch,
                 new_apagenum, new_vpagenum, erase=True):
        """Change the video mode, colourburst, visible or active page."""
        # preserve memory if erase==0; don't distingush erase==1 and erase==2
        save_mem = None
        if (not erase and self.mode.video_segment == spec.video_segment):
            save_mem = self.mode.get_all_memory(self)
        # reset palette happens even if the SCREEN call fails
        self.video.set_cga4_palette(1)
        # if the new mode has fewer pages than current vpage/apage,
        # illegal fn call before anything happens.
        # signal the signals to change the screen resolution
        if (not spec or new_apagenum >= spec.num_pages or new_vpagenum >= spec.num_pages):
            raise error.BASICError(error.IFC)
        # illegal fn call if we don't have a font for this mode
        self.text_screen.check_font_available(spec)
        # if we made it here we're ready to commit to the new mode
        self.queues.video.put(signals.Event(signals.VIDEO_SET_MODE, (spec,)))
        # switching to another text mode (width-only change)
        width_only = (self.mode.is_text_mode and spec.is_text_mode)
        # attribute and border persist on width-only change
        # otherwise start with black border and default attr
        if (
                not width_only or self.apagenum != new_apagenum or self.vpagenum != new_vpagenum
                or self.colorswitch != new_colorswitch
            ):
            self.attr = spec.attr
        if (not width_only and spec.name != self.mode.name):
            self.set_border(0)
        # set the screen mode parameters
        self.mode, self._mode_nr = spec, new_mode
        # initialise the palette
        self.palette.init_mode(self.mode)
        # set the colorswitch
        self._init_mode_colorburst(new_colorswitch)
        # initialise pixel buffers
        if not self.mode.is_text_mode:
            self.pixels = PixelBuffer(
                self.mode.pixel_width, self.mode.pixel_height,
                self.mode.num_pages, self.mode.bitsperpixel
            )
        else:
            self.pixels = None
        # set active page & visible page, counting from 0.
        self.set_page(new_vpagenum, new_apagenum)
        # initialise text screen
        # NOTE this requires active page to have beet set!
        self.text_screen.init_mode(self.mode, self.pixels, self.attr, new_vpagenum, new_apagenum)
        # restore emulated video memory in new mode
        if save_mem:
            self.mode.set_all_memory(self, save_mem)
        # center graphics cursor, reset window, etc.
        self.drawing.init_mode(self.mode, self.text_screen.text, self.pixels)
        self.drawing.set_attr(self.attr)

    def set_width(self, to_width):
        """Set the character width of the screen, reset pages and change modes."""
        # raise an error if the width value doesn't make sense
        if to_width not in self.video.get_allowed_widths():
            raise error.BASICError(error.IFC)
        # if we're currently at that width, do nothing
        if to_width == self.mode.width:
            return
        elif to_width == 20:
            self.screen(3, None, 0, 0)
        elif self.mode.is_text_mode:
            self.screen(0, None, 0, 0, new_width=to_width)
        elif to_width == 40:
            if self.mode.name == '640x200x2':
                self.screen(1, None, 0, 0)
            elif self.mode.name == '160x200x16':
                self.screen(1, None, 0, 0)
            elif self.mode.name == '640x200x4':
                self.screen(5, None, 0, 0)
            elif self.mode.name == '640x200x16':
                self.screen(7, None, 0, 0)
            elif self.mode.name == '640x350x16':
                # screen 9 switches to screen 1 (not 7) on WIDTH 40
                self.screen(1, None, 0, 0)
        elif to_width == 80:
            if self.mode.name == '320x200x4':
                self.screen(2, None, 0, 0)
            elif self.mode.name == '160x200x16':
                self.screen(2, None, 0, 0)
            elif self.mode.name == '320x200x4pcjr':
                self.screen(2, None, 0, 0)
            elif self.mode.name == '320x200x16pcjr':
                self.screen(6, None, 0, 0)
            elif self.mode.name == '320x200x16':
                self.screen(8, None, 0, 0)
        else:
            raise error.BASICError(error.IFC)

    def _init_mode_colorburst(self, new_colorswitch):
        """Initialise colorburst settings for new screen mode and colorswitch."""
        self.colorswitch = new_colorswitch
        # in screen 0, 1, set colorburst (not in SCREEN 2!)
        if self.mode.is_text_mode:
            self.set_colorburst(new_colorswitch)
        elif self.mode.name == '320x200x4':
            self.set_colorburst(not new_colorswitch)
        elif self.mode.name == '640x200x2':
            self.set_colorburst(False)

    def set_colorburst(self, on=True):
        """Set the composite colorburst bit."""
        colorburst = self.video.set_colorburst(on, is_cga=(self.mode.name == '320x200x4'))
        # reset the palette to reflect the new mono or mode-5 situation
        # this sends the signal to the interface as well
        self.palette.init_mode(self.mode)
        # don't try composite unless our video card supports it
        if self.capabilities in modes.COMPOSITE:
            composite_artifacts = (
                colorburst and self.video.monitor == 'composite' and
                (not self.mode.is_text_mode) and self.mode.supports_artifacts
            )
            if composite_artifacts:
                # set a composite palette
                self.queues.video.put(signals.Event(
                    signals.VIDEO_SET_PALETTE,
                    (modes.COMPOSITE[self.capabilities], None, (4, self.mode.bitsperpixel))
                ))
            else:
                # set normal palette
                self.queues.video.put(signals.Event(
                    signals.VIDEO_SET_PALETTE,
                    (self.palette.rgb_palette, self.palette.rgb_palette1, None)
                ))


    def set_video_memory_size(self, new_size):
        """Change the amount of memory available to the video card."""
        # redefine number of available video pages
        self.video.prepare_modes(new_size)
        # text screen modes don't depend on video memory size
        if self._mode_nr == 0:
            return
        # check if we need to drop out of our current mode
        page = max(self.vpagenum, self.apagenum)
        # reload max number of pages; do we fit? if not, drop to text
        new_mode = self.video.get_mode(self._mode_nr)
        if (page >= new_mode.num_pages):
            self.screen(0, 0, 0, 0, force_reset=True)
        else:
            self.mode = new_mode

    def rebuild(self):
        """Completely resubmit the screen to the interface."""
        # set the screen mode
        self.queues.video.put(signals.Event(signals.VIDEO_SET_MODE, (self.mode,)))
        # set the visible and active pages
        self.queues.video.put(signals.Event(signals.VIDEO_SET_PAGE, (self.vpagenum, self.apagenum)))
        # rebuild palette
        self.palette.set_all(self.palette.palette, check_mode=False)
        # set the border
        fore, _, _, _ = self.mode.split_attr(self.border_attr)
        self.queues.video.put(signals.Event(signals.VIDEO_SET_BORDER_ATTR, (fore,)))
        self.text_screen.rebuild()


    ###########################################################################

    def set_page(self, new_vpagenum, new_apagenum):
        """Set active page & visible page, counting from 0."""
        if new_vpagenum is None:
            new_vpagenum = self.vpagenum
        if new_apagenum is None:
            new_apagenum = self.apagenum
        if (new_vpagenum >= self.mode.num_pages or new_apagenum >= self.mode.num_pages):
            raise error.BASICError(error.IFC)
        self.vpagenum = new_vpagenum
        self.apagenum = new_apagenum
        self.drawing.set_page(new_apagenum)
        self.text_screen.set_page(new_vpagenum, new_apagenum)
        self.queues.video.put(signals.Event(signals.VIDEO_SET_PAGE, (new_vpagenum, new_apagenum)))

    def set_attr(self, attr):
        """Set the default attribute."""
        self.attr = attr
        self.drawing.set_attr(attr)
        self.text_screen.set_attr(attr)
        if not self.mode.is_text_mode and self.mode.cursor_index is None:
            fore, _, _, _ = self.mode.split_attr(attr)
            self.queues.video.put(signals.Event(signals.VIDEO_SET_CURSOR_ATTR, (fore,)))

    def set_border(self, attr):
        """Set the border attribute."""
        self.border_attr = attr
        fore, _, _, _ = self.mode.split_attr(attr)
        self.queues.video.put(signals.Event(signals.VIDEO_SET_BORDER_ATTR, (fore,)))

    ###########################################################################
    # memory operations

    def get_memory(self, addr, num_bytes):
        """Retrieve bytes from video memory."""
        return self.mode.get_memory(self, addr, num_bytes)

    def set_memory(self, addr, bytestr):
        """Set bytes in video memory."""
        self.mode.set_memory(self, addr, bytestr)

    ###########################################################################
    # callbacks

    # SCREEN statement
    # Colorswitch parameter
    #   SCREEN 0,0 - mono on composite, color on RGB
    #   SCREEN 0,1 - color (colorburst=True)
    #   SCREEN 1,0 - color (colorburst=True)
    #   SCREEN 1,1 - mono on composite, mode 5 on RGB
    # default colorswitch:
    #   SCREEN 0 = SCREEN 0,0 (pcjr)
    #   SCREEN 0 = SCREEN 0,1 (tandy, cga, ega, vga, ..)
    #   SCREEN 1 = SCREEN 1,0 (pcjr, tandy)
    #   SCREEN 1 = SCREEN 1,1 (cga, ega, vga, ...)
    # colorswitch is NOT preserved between screens when unspecified
    # colorswitch is NOT the same as colorburst (opposite on screen 1)
    #
    # Erase parameter:
    #   tells basic how much video memory to erase
    #   0: do not erase video memory
    #   1: (default) erase old and new page if screen or width changes
    #   2: erase all video memory if screen or width changes
    #   -> we're not distinguishing between 1 and 2 here

    def screen_(self, args):
        """SCREEN: change the video mode, colourburst, visible or active page."""
        # in GW, screen 0,0,0,0,0,0 raises error after changing the palette
        # this raises error before
        mode, colorswitch, apagenum, vpagenum = (
            None if arg is None else values.to_int(arg)
            for _, arg in zip(range(4), args)
        )
        # if any parameter not in [0,255], error 5 without doing anything
        # if the parameters are outside narrow ranges
        # (e.g. not implemented screen mode, pagenum beyond max)
        # then the error is only raised after changing the palette.
        error.range_check(0, 255, mode, colorswitch, apagenum, vpagenum)
        if self.capabilities == 'tandy':
            error.range_check(0, 1, colorswitch)
        erase = next(args)
        if erase is not None:
            erase = values.to_int(erase)
            error.range_check(0, 2, erase)
        list(args)
        if erase is not None:
            # erase can only be set on pcjr/tandy 5-argument syntax
            if self.capabilities not in ('pcjr', 'tandy'):
                raise error.BASICError(error.IFC)
        else:
            erase = 1
        self.screen(mode, colorswitch, apagenum, vpagenum, erase)

    def pcopy_(self, args):
        """Copy source to destination page."""
        src = values.to_int(next(args))
        error.range_check(0, self.mode.num_pages-1, src)
        dst = values.to_int(next(args))
        list(args)
        error.range_check(0, self.mode.num_pages-1, dst)
        self.text_screen.text.copy_page(src, dst)
        if not self.mode.is_text_mode:
            self.pixels.copy_page(src, dst)
        self.queues.video.put(signals.Event(signals.VIDEO_COPY_PAGE, (src, dst)))

    def color_(self, args):
        """COLOR: set colour attributes."""
        args = list(args)
        error.throw_if(len(args) > 3)
        args += [None] * (3 - len(args))
        fore, back, bord = args
        if fore is None:
            fore = (self.attr >> 7) * 0x10 + (self.attr & 0xf)
        else:
            fore = values.to_int(fore)
        if back is not None:
            back = back and values.to_int(back)
        if bord is not None:
            bord = bord and values.to_int(bord)
        if self.mode.name in ('640x200x2', '720x348x2'):
            # screen 2; hercules: illegal fn call
            raise error.BASICError(error.IFC)
        elif self.mode.name == '320x200x4':
            self._color_mode_1(fore, back, bord)
        else:
            self._color_other_modes(fore, back, bord)

    def _color_mode_1(self, back, pal, override):
        """Helper function for COLOR in SCREEN 1."""
        back = self.palette.get_entry(0) if back is None else back
        if override is not None:
            # uses last entry as palette if given
            pal = override
        error.range_check(0, 255, back)
        if pal is not None:
            error.range_check(0, 255, pal)
            self.video.set_cga4_palette(pal % 2)
            palette = list(self.mode.palette)
            palette[0] = back & 0xf
            # cga palette 0: 0,2,4,6    hi 0, 10, 12, 14
            # cga palette 1: 0,3,5,7 (Black, Ugh, Yuck, Bleah), hi: 0, 11,13,15
            self.palette.set_all(palette, check_mode=False)
        else:
            self.palette.set_entry(0, back & 0xf, check_mode=False)

    def _color_other_modes(self, fore, back, bord):
        """Helper function for COLOR in modes other than SCREEN 1."""
        mode = self.mode
        if back is None:
            # graphics mode bg is always 0; sets palette instead
            if mode.is_text_mode:
                back = (self.attr >> 4) & 0x7
            else:
                back = self.palette.get_entry(0)
        # for screens other than 1, no distinction between 3rd parm zero and not supplied
        bord = bord or 0
        error.range_check(0, 255, bord)
        if mode.is_text_mode:
            error.range_check(0, mode.num_attr-1, fore)
            error.range_check(0, 15, back, bord)
            self.set_attr(((0x8 if (fore > 0xf) else 0x0) + (back & 0x7))*0x10 + (fore & 0xf))
            self.set_border(bord)
        elif mode.name in (
                '160x200x16', '320x200x4pcjr', '320x200x16pcjr'
                '640x200x4', '320x200x16', '640x200x16'
            ):
            error.range_check(1, mode.num_attr-1, fore)
            error.range_check(0, mode.num_attr-1, back)
            self.set_attr(fore)
            # in screen 7 and 8, only low intensity palette is used.
            self.palette.set_entry(0, back % 8, check_mode=False)
        elif mode.name in ('640x350x16', '640x350x4'):
            error.range_check(1, mode.num_attr-1, fore)
            error.range_check(0, len(mode.colours)-1, back)
            self.set_attr(fore)
            self.palette.set_entry(0, back, check_mode=False)
        elif mode.name == '640x400x2':
            error.range_check(0, len(mode.colours)-1, fore)
            if back != 0:
                raise error.BASICError(error.IFC)
            self.palette.set_entry(1, fore, check_mode=False)

    def cls_(self, args):
        """CLS: clear the screen."""
        val = next(args)
        if val is not None:
            val = values.to_int(val)
            # tandy gives illegal function call on CLS number
            error.throw_if(self.capabilities == 'tandy')
            error.range_check(0, 2, val)
        else:
            if self.drawing.graph_view.is_set():
                val = 1
            elif self.text_screen.scroll_area.active:
                val = 2
            else:
                val = 0
        list(args)
        # cls is only executed if no errors have occurred
        if val == 0:
            self.text_screen.clear()
            self.text_screen.bottom_bar.redraw(self.text_screen)
            self.drawing.reset()
        elif val == 1:
            # clear the graphics viewport
            if not self.mode.is_text_mode:
                self.drawing.fill_rect(*self.drawing.graph_view.get(), index=(self.attr >> 4) & 0x7)
            self.drawing.reset()
        elif val == 2:
            self.text_screen.clear_view()



"""
PC-BASIC - state.py
Support for pickling emulator state

(c) 2014--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

try:
    import cPickle as pickle
except ImportError:
    import pickle

import os
import io
import sys
import zlib
import struct
import logging
from contextlib import contextmanager

from .metadata import VERSION
from .compat import PY2, copyreg, stdout, stdin

# session file header
HEADER_FORMAT = '<LIIIII'
HEADER_KEYS = [
    'checksum', 'format_version', 'python_major', 'python_minor', 'pcbasic_major', 'pcbasic_minor'
]
HEADER = {
    # increment this if we change the format of the session file
    'format_version': 2,
    'python_major': sys.version_info.major,
    'python_minor': sys.version_info.minor,
    'pcbasic_major': int(VERSION.split(u'.')[0]),
    'pcbasic_minor': int(VERSION.split(u'.')[1]),
}

@contextmanager
def manage_state(session, state_file, do_resume):
    """Resume a session if requested; save upon exit"""
    if do_resume and state_file:
        try:
            session = load_session(state_file).attach(session.interface)
        except Exception as e:
            # if we were told to resume but can't, give up
            logging.fatal('Failed to resume session from %s: %s', state_file, e)
            sys.exit(1)
    try:
        yield session
    finally:
        if state_file:
            try:
                save_session(session, state_file)
            except Exception as e:
                logging.error('Failed to save session to %s: %s', state_file, e)


def unpickle_file(name, mode, pos):
    """Unpickle a file object."""
    if name is None:
        if mode in ('r', 'rb'):
            return sys.stdin
        else:
            return sys.stdout
    try:
        if 'w' in mode and pos > 0:
            # preserve existing contents of writable file
            with io.open(name, 'rb') as f:
                buf = f.read(pos)
            f = io.open(name, mode)
            f.write(buf)
        else:
            f = io.open(name, mode)
            if pos > 0:
                f.seek(pos)
    except IOError:
        logging.warning('Could not re-open file %s. Replacing with null file.', name)
        f = io.open(os.devnull, mode)
    return f

def pickle_file(f):
    """Pickle a file object."""
    if f in (sys.stdout, sys.stdin, stdout, stdin, stdout.buffer, stdin.buffer):
        return unpickle_file, (None, f.mode, -1)
    try:
        return unpickle_file, (f.name, f.mode, f.tell())
    except IOError:
        # not seekable
        return unpickle_file, (f.name, f.mode, -1)

# register the picklers for file and cStringIO
if PY2:
    copyreg.pickle(file, pickle_file)  # pylint: disable=undefined-variable
copyreg.pickle(io.BufferedReader, pickle_file)
copyreg.pickle(io.BufferedWriter, pickle_file)
copyreg.pickle(io.TextIOWrapper, pickle_file)
copyreg.pickle(io.BufferedRandom, pickle_file)


def load_session(state_file):
    """Read state from a compressed pickle."""
    with open(state_file, 'rb') as in_file:
        header = in_file.read(struct.calcsize(HEADER_FORMAT))
        blob = in_file.read()
    # mask checksum to deal with different signs on Py2/Py3
    # see https://docs.python.org/3.5/library/zlib.html#zlib.crc32
    checksum = zlib.crc32(blob) & 0xffffffff
    try:
        header_dict = dict(zip(HEADER_KEYS, struct.unpack(HEADER_FORMAT, header)))
    except struct.error:
        raise ValueError('session file header corrupted')
    # check blob integrity
    if checksum != header_dict['checksum']:
        raise ValueError('session file corrupted')
    if (
            HEADER['python_major'] != header_dict['python_major']
            or HEADER['python_minor'] != header_dict['python_minor']
        ):
        raise ValueError('session file stored with different Python version')
    if (
            HEADER['pcbasic_major'] != header_dict['pcbasic_major']
            or HEADER['pcbasic_minor'] != header_dict['pcbasic_minor']
        ):
        raise ValueError('session file stored with different PC-BASIC version')
    session = pickle.loads(zlib.decompress(blob))
    return session

def save_session(obj, state_file):
    """Write state to a compressed pickle."""
    blob = zlib.compress(pickle.dumps(obj, pickle.HIGHEST_PROTOCOL))
    checksum = zlib.crc32(blob) & 0xffffffff
    header_dict = dict(checksum=checksum, **HEADER)
    header = struct.pack(HEADER_FORMAT, *(header_dict[_key] for _key in HEADER_KEYS))
    with open(state_file, 'wb') as out_file:
        out_file.write(header)
        out_file.write(blob)



"""
PC-BASIC - inputmethods.py
Pen and joystick handling

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import datetime

from ..base import error
from ..base import scancode
from ..base import tokens as tk
from ..base import signals
from .. import values


###############################################################################
# clipboard copy & print screen handler

# clipboard copy & print screen are special cases:
# they to handle an input signal, read the screen
# and write the text to an output queue or file
# independently of what BASIC is doing

class ScreenCopyHandler(object):
    """Event handler for clipboard copy and print screen."""

    def __init__(self, screen, lpt1_file):
        """Initialise copy handler."""
        self._screen = screen
        self._lpt1_file = lpt1_file

    def check_input(self, signal):
        """Handle input signals."""
        if signal.event_type == signals.CLIP_COPY:
            self._screen.copy_clipboard(*signal.params)
            return True
        elif signal.event_type == signals.KEYB_DOWN:
            c, scan, mod = signal.params
            if scan == scancode.PRINT and (
                    scancode.LSHIFT in mod or scancode.RSHIFT in mod):
                # shift+printscreen triggers a print screen
                self._screen.print_screen(self._lpt1_file)
                return True
        return False


###############################################################################
# light pen

class Pen(object):
    """Light pen support."""

    def __init__(self):
        """Initialise light pen."""
        self._is_down = False
        self._pos = 0, 0
        # signal pen has been down for PEN polls in pen_()
        self._was_down = False
        self._down_pos = (0, 0)

    def check_input(self, signal):
        """Handle pen-related input signals."""
        if signal.event_type == signals.PEN_DOWN:
            self._down(*signal.params)
        elif signal.event_type == signals.PEN_UP:
            self._up()
        elif signal.event_type == signals.PEN_MOVED:
            self._moved(*signal.params)
        else:
            return False
        return True

    def _down(self, x, y):
        """Report a pen-down event at graphical x,y """
        # TRUE until polled
        self._was_down = True
        # TRUE until pen up
        self._is_down = True
        self._down_pos = x, y

    def _up(self):
        """Report a pen-up event at graphical x,y """
        self._is_down = False

    def _moved(self, x, y):
        """Report a pen-move event at graphical x,y """
        self._pos = x, y

    def poll(self, fn, enabled, screen):
        """PEN: poll the light pen."""
        fn = values.to_int(fn)
        error.range_check(0, 9, fn)
        posx, posy = self._pos
        if fn == 0:
            pen_down_old, self._was_down = self._was_down, False
            pen = -1 if pen_down_old else 0
        elif fn == 1:
            pen = self._down_pos[0]
        elif fn == 2:
            pen = self._down_pos[1]
        elif fn == 3:
            pen = -1 if self._is_down else 0
        elif fn == 4:
            pen = posx
        elif fn == 5:
            pen = posy
        elif fn == 6:
            pen = 1 + self._down_pos[1] // screen.mode.font_height
        elif fn == 7:
            pen = 1 + self._down_pos[0] // screen.mode.font_width
        elif fn == 8:
            pen = 1 + posy // screen.mode.font_height
        elif fn == 9:
            pen = 1 + posx // screen.mode.font_width
        if not enabled:
            # should return 0 or char pos 1 if PEN not ON
            pen = 1 if fn >= 6 else 0
        return pen


###############################################################################
# joysticks


class Stick(object):
    """Joystick support."""

    def __init__(self, values):
        """Initialise joysticks."""
        self._values = values
        self.is_firing = [[False, False], [False, False]]
        # axis 0--255; 128 is mid but reports 0, not 128 if no joysticks present
        self.axis = [[0, 0], [0, 0]]
        self.is_on = False
        self._was_fired = [[False, False], [False, False]]
        # timer for reading game port
        self._out_time = self._decay_timer()

    def check_input(self, signal):
        """Handle joystick-related input signals."""
        if signal.event_type == signals.STICK_DOWN:
            self._down(*signal.params)
        elif signal.event_type == signals.STICK_UP:
            self._up(*signal.params)
        elif signal.event_type == signals.STICK_MOVED:
            self._moved(*signal.params)
        else:
            return False
        return True

    def _down(self, joy, button):
        """Report a joystick button down event."""
        try:
            self._was_fired[joy][button] = True
            self.is_firing[joy][button] = True
        except IndexError:
            # ignore any joysticks/axes beyond the 2x2 supported by BASIC
            pass

    def _up(self, joy, button):
        """Report a joystick button up event."""
        try:
            self.is_firing[joy][button] = False
        except IndexError:
            # ignore any joysticks/axes beyond the 2x2 supported by BASIC
            pass

    def _moved(self, joy, axis, value):
        """Report a joystick axis move."""
        try:
            self.axis[joy][axis] = value
        except IndexError:
            # ignore any joysticks/axes beyond the 2x2 supported by BASIC
            pass

    def strig_statement_(self, args):
        """Switch joystick handling on or off."""
        on, = args
        self.is_on = (on == tk.ON)

    def stick_(self, args):
        """STICK: poll the joystick axes."""
        fn, = args
        fn = values.to_int(fn)
        error.range_check(0, 3, fn)
        joy, axis = fn // 2, fn % 2
        try:
            result = self.axis[joy][axis]
        except IndexError:
            # ignore any joysticks/axes beyond the 2x2 supported by BASIC
            result = 0
        return self._values.new_integer().from_int(result)

    def strig_(self, args):
        """STRIG: poll the joystick fire button."""
        fn, = args
        fn = values.to_int(fn)
        error.range_check(0, 7, fn)
        # 0,1 -> [0][0] 2,3 -> [0][1]  4,5-> [1][0]  6,7 -> [1][1]
        joy, trig = fn // 4, (fn//2) % 2
        if fn % 2 == 0:
            # has been fired
            stick_was_trig = self._was_fired[joy][trig]
            self._was_fired[joy][trig] = False
            result = -1 if stick_was_trig else 0
        else:
            # is currently firing
            result = -1 if self.is_firing[joy][trig] else 0
        return self._values.new_integer().from_int(result)

    def decay(self):
        """Return time since last game port reset."""
        return (self._decay_timer() - self._out_time) % 86400000

    def reset_decay(self):
        """Reset game port."""
        self._out_time = self._decay_timer()

    def _decay_timer(self):
        """Millisecond timer for game port decay."""
        now = datetime.datetime.now()
        return now.second*1000 + now.microsecond//1000



"""
PC-BASIC - GW-BASIC/BASICA/Cartridge BASIC compatible interpreter

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import os as _os

# __path__ hack to ensure os.chdir does not break intra-package imports
# which they do because the package __path__ is given relative to cwd
# at least if run with python -m package
__path__ = [_os.path.abspath(_e) for _e in __path__]

from .metadata import VERSION as __version__
from .basic.api import Session, codepage, font
from .main import run, main



"""
PC-BASIC - textbase.py
Text screen helper classes

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from ...compat import iterchar
from ..base import signals
from ..base import error
from .. import values


#######################################################################################
# function key macro guide

class BottomBar(object):
    """Key guide bar at bottom line."""

    def __init__(self):
        """Initialise bottom bar."""
        # use 80 here independent of screen width
        # we store everything in a buffer and only show what fits
        self.clear()
        self.visible = False

    def clear(self):
        """Clear the contents."""
        self._contents = [(b' ', 0)] * 80

    def write(self, s, col, reverse):
        """Write chars on virtual bottom bar."""
        for i, c in enumerate(iterchar(s)):
            self._contents[col + i] = (c, reverse)

    def show(self, on, screen):
        """Switch bottom bar visibility."""
        # tandy can have VIEW PRINT 1 to 25, should raise IFC in that case
        error.throw_if(on and screen.scroll_area.bottom == screen.mode.height)
        self.visible, was_visible = on, self.visible
        if self.visible != was_visible:
            self.redraw(screen)

    def redraw(self, screen):
        """Redraw bottom bar if visible, clear if not."""
        key_row = screen.mode.height
        # Keys will only be visible on the active page at which KEY ON was given,
        # and only deleted on page at which KEY OFF given.
        screen.clear_rows(key_row, key_row)
        if not screen.mode.is_text_mode:
            reverse_attr = screen.attr
        elif (screen.attr >> 4) & 0x7 == 0:
            reverse_attr = 0x70
        else:
            reverse_attr = 0x07
        if self.visible:
            # always show only complete 8-character cells
            # this matters on pcjr/tandy width=20 mode
            for i in range((screen.mode.width//8) * 8):
                c, reverse = self._contents[i]
                a = reverse_attr if reverse else screen.attr
                screen.put_char_attr(screen.apagenum, key_row, i+1, c, a)
            screen.text.pages[screen.apagenum].row[key_row-1].end = screen.mode.width


#######################################################################################
# cursor

class Cursor(object):
    """Manage the cursor."""

    def __init__(self, queues, mode, capabilities):
        """Initialise the cursor."""
        self._queues = queues
        self._mode = mode
        self._capabilities = capabilities
        # are we in parse mode? invisible unless visible_run is True
        self.default_visible = True
        # cursor visible in parse mode? user override
        self.visible_run = False
        # cursor shape
        self.from_line = 0
        self.to_line = 0
        self.width = self._mode.font_width
        self._height = self._mode.font_height

    def init_mode(self, mode, attr):
        """Change the cursor for a new screen mode."""
        self._mode = mode
        self.width = mode.font_width
        self._height = mode.font_height
        # set the cursor attribute
        if not mode.is_text_mode:
            fore, _, _, _ = mode.split_attr(mode.cursor_index or attr)
            self._queues.video.put(signals.Event(signals.VIDEO_SET_CURSOR_ATTR, (fore,)))
        # cursor width starts out as single char
        self.set_default_shape(True)
        self.reset_visibility()

    def reset_attr(self, new_attr):
        """Set the text cursor attribute."""
        if self._mode.is_text_mode:
            self._queues.video.put(signals.Event(signals.VIDEO_SET_CURSOR_ATTR, (new_attr,)))

    def show(self, do_show):
        """Force cursor to be visible/invisible."""
        self._queues.video.put(signals.Event(signals.VIDEO_SHOW_CURSOR, (do_show,)))

    def set_visibility(self, visible_run):
        """Set cursor visibility when a program is being run."""
        self.visible_run = visible_run
        self.reset_visibility()

    def reset_visibility(self):
        """Set cursor visibility to its default state."""
        # visible if in interactive mode and invisible when a program is being run
        visible = self.default_visible
        # unless forced to be visible
        # in graphics mode, we can't force the cursor to be visible on execute.
        if self._mode.is_text_mode:
            visible = visible or self.visible_run
        self._queues.video.put(signals.Event(signals.VIDEO_SHOW_CURSOR, (visible,)))

    def set_shape(self, from_line, to_line):
        """Set the cursor shape."""
        # A block from from_line to to_line in 8-line modes.
        # Use compatibility algo in higher resolutions
        fx, fy = self.width, self._height
        # do all text modes with >8 pixels have an ega-cursor?
        if self._capabilities in (
                'ega', 'mda', 'ega_mono', 'vga', 'olivetti', 'hercules'):
            # odd treatment of cursors on EGA machines,
            # presumably for backward compatibility
            # the following algorithm is based on DOSBox source int10_char.cpp
            #     INT10_SetCursorShape(Bit8u first,Bit8u last)
            max_line = fy - 1
            if from_line & 0xe0 == 0 and to_line & 0xe0 == 0:
                if (to_line < from_line):
                    # invisible only if to_line is zero and to_line < from_line
                    if to_line != 0:
                        # block shape from *to_line* to end
                        from_line = to_line
                        to_line = max_line
                elif ((from_line | to_line) >= max_line or
                            to_line != max_line-1 or from_line != max_line):
                    if to_line > 3:
                        if from_line+2 < to_line:
                            if from_line > 2:
                                from_line = (max_line+1) // 2
                            to_line = max_line
                        else:
                            from_line = from_line - to_line + max_line
                            to_line = max_line
                            if max_line > 0xc:
                                from_line -= 1
                                to_line -= 1
        self.from_line = max(0, min(from_line, fy-1))
        self.to_line = max(0, min(to_line, fy-1))
        self._queues.video.put(signals.Event(signals.VIDEO_SET_CURSOR_SHAPE,
                            (self.width, fy, self.from_line, self.to_line)))

    def set_default_shape(self, overwrite_shape):
        """Set the cursor to one of two default shapes."""
        if overwrite_shape:
            if not self._mode.is_text_mode:
                # always a block cursor in graphics mode
                self.set_shape(0, self._height-1)
            elif self._capabilities == 'ega':
                # EGA cursor is on second last line
                self.set_shape(self._height-2, self._height-2)
            elif self._height == 9:
                # Tandy 9-pixel fonts; cursor on 8th
                self.set_shape(self._height-2, self._height-2)
            else:
                # other cards have cursor on last line
                self.set_shape(self._height-1, self._height-1)
        else:
            # half-block cursor for insert
            self.set_shape(self._height//2, self._height-1)

    def set_width(self, num_chars):
        """Set the cursor with to num_chars characters."""
        new_width = num_chars * self._mode.font_width
        # update cursor shape to new width if necessary
        if new_width != self.width:
            self.width = new_width
            self._queues.video.put(signals.Event(signals.VIDEO_SET_CURSOR_SHAPE,
                    (self.width, self._height, self.from_line, self.to_line)))


###############################################################################
# text viewport / scroll area

class ScrollArea(object):
    """Text viewport / scroll area."""

    def __init__(self, mode):
        """Initialise the scroll area."""
        self._height = mode.height
        self.unset()

    def init_mode(self, mode):
        """Initialise the scroll area for new screen mode."""
        self._height = mode.height
        if self._bottom == self._height:
            # tandy/pcjr special case: VIEW PRINT to 25 is preserved
            self.set(1, self._height)
        else:
            self.unset()

    @property
    def active(self):
        """A viewport has been set."""
        return self._active

    @property
    def bounds(self):
        """Return viewport bounds."""
        return self._top, self._bottom

    @property
    def top(self):
        """Return viewport top bound."""
        return self._top

    @property
    def bottom(self):
        """Return viewport bottom bound."""
        return self._bottom

    def set(self, start, stop):
        """Set the scroll area."""
        self._active = True
        # _top and _bottom are inclusive and count rows from 1
        self._top = start
        self._bottom = stop
        #  need this:
        #set_pos(start, 1)
        #  or this:
        #self.overflow = False
        #self._move_cursor(start, 1)

    def unset(self):
        """Unset scroll area."""
        # there is only one VIEW PRINT setting across all pages.
        # scroll area normally excludes the bottom bar
        self.set(1, self._height - 1)
        self._active = False



"""
PC-BASIC - text.py
Text-buffer operations

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import logging
import binascii

class TextRow(object):
    """Buffer for a single row of the screen."""

    def __init__(self, attr, width, conv, dbcs_enabled):
        """Set up screen row empty and unwrapped."""
        self.width = width
        self.clear(attr)
        # line continues on next row (either LF or word wrap happened)
        self.wrap = False
        self._dbcs_enabled = dbcs_enabled
        self._conv = conv

    def clear(self, attr):
        """Clear the screen row buffer. Leave wrap untouched."""
        # screen buffer, initialised to spaces
        self.buf = [(b' ', attr)] * self.width
        # character is part of double width char; 0 = no; 1 = lead, 2 = trail
        self.double = [0] * self.width
        # last non-whitespace character
        self.end = 0

    def clear_from(self, scol, attr):
        """Clear characters from given position till end of row."""
        self.buf = self.buf[:scol-1] + [(b' ', attr)] * (self.width - scol + 1)
        self.double = self.double[:scol-1] + [0] * (self.width - scol + 1)
        self.end = min(self.end, scol-1)

    def put_char_attr(self, col, c, attr):
        """Put a byte to the screen, reinterpreting SBCS and DBCS as necessary."""
        # update the screen buffer
        self.buf[col-1] = (c, attr)
        self.double[col-1] = 0
        # for sbcs codepages we're done now
        if not self._dbcs_enabled:
            return col, col
        # mark out replaced char and changed following dbcs characters to be redrawn
        sequences = self._conv.mark(b''.join(entry[0] for entry in self.buf), flush=True)
        flags = ((0,) if len(seq) == 1 else (1, 2) for seq in sequences)
        old_double = self.double
        self.double = [entry for flag in flags for entry in flag]
        # find the first and last changed columns, to be able to redraw
        diff = [old != new for old, new in zip(old_double, self.double)]
        if True in diff:
            start, stop = diff.index(True) + 1, len(diff) - diff[::-1].index(True)
        else:
            start, stop = col, col
        # if the tail byte has changed, the lead byte needs to be redrawn as well
        if self.double[start-1] == 2:
            start -= 1
        return min(col, start), max(col, stop)


class TextPage(object):
    """Buffer for a screen page."""

    def __init__(self, attr, width, height, conv, dbcs_enabled):
        """Initialise the screen buffer to given dimensions."""
        self.row = [TextRow(attr, width, conv, dbcs_enabled) for _ in range(height)]
        self.width = width
        self.height = height


class TextBuffer(object):
    """Buffer for text on all screen pages."""

    def __init__(self, attr, width, height, num_pages, codepage, do_fullwidth):
        """Initialise the screen buffer to given pages and dimensions."""
        self._dbcs_enabled = codepage.dbcs and do_fullwidth
        self._conv = codepage.get_converter(preserve=b'')
        self.pages = [
            TextPage(attr, width, height, self._conv, self._dbcs_enabled)
            for _ in range(num_pages)
        ]
        self.width = width
        self.height = height

    def __repr__(self):
        """Return an ascii representation of the screen buffer (for debugging)."""
        horiz_bar = ('   +' + '-' * self.width + '+')
        row_strs = []
        for num, page in enumerate(self.pages):
            lastwrap = False
            row_strs.append(horiz_bar)
            for i, row in enumerate(page.row):
                # convert non-ascii bytes to \x81 etc
                # dbcs is encoded as double char in left column, '' in right
                rowbytes = (_pair[0] for _pair in row.buf)
                # replace non-ascii with ? - this is not ideal but
                # for python2 we need to stick to ascii-128 so implicit conversion to bytes works
                # and for python3 we must use unicode
                # and backslashreplace messes up the output width...
                rowstr = ''.join(
                    _char.decode('ascii', 'replace').replace(u'\ufffd', u'?')
                    for _char in rowbytes
                )
                left = '\\' if lastwrap else '|'
                right = '\\' if row.wrap else '|'
                row_strs.append('{0:2} {1}{2}{3} {4:2}'.format(
                    i, left, rowstr, right, row.end,
                ))
                lastwrap = row.wrap
            row_strs.append(horiz_bar)
        return '\n'.join(row_strs)

    def copy_page(self, src, dst):
        """Copy source to destination page."""
        for x in range(self.height):
            dstrow = self.pages[dst].row[x]
            srcrow = self.pages[src].row[x]
            dstrow.buf[:] = srcrow.buf[:]
            dstrow.end = srcrow.end
            dstrow.wrap = srcrow.wrap

    def clear_area(self, pagenum, row0, col0, row1, col1, attr):
        """Clear a rectangular area of the screen."""
        for r in range(row0-1, row1):
            self.pages[pagenum].row[r].buf[col0-1:col1] = [(b' ', attr)] * (col1 - col0 + 1)

    def put_char_attr(self, pagenum, row, col, c, attr):
        """Put a byte to the screen, reinterpreting SBCS and DBCS as necessary."""
        assert isinstance(c, bytes), type(c)
        return self.pages[pagenum].row[row-1].put_char_attr(col, c, attr)

    def scroll_up(self, pagenum, from_line, bottom, attr):
        """Scroll up."""
        self.pages[pagenum].row.insert(
            bottom, TextRow(attr, self.width, self._conv, self._dbcs_enabled)
        )
        del self.pages[pagenum].row[from_line-1]

    def scroll_down(self, pagenum, from_line, bottom, attr):
        """Scroll down."""
        self.pages[pagenum].row.insert(
            from_line - 1, TextRow(attr, self.width, self._conv, self._dbcs_enabled)
        )
        # if we were already a wrapping row, make sure the new empty row wraps
        if self.pages[pagenum].row[from_line-2].wrap:
            self.pages[pagenum].row[from_line-1].wrap = True
        del self.pages[pagenum].row[bottom-1]

    def get_char(self, pagenum, row, col):
        """Retrieve a byte from the screen (SBCS or DBCS half-char)."""
        return ord(self.pages[pagenum].row[row-1].buf[col-1][0])

    def get_attr(self, pagenum, row, col):
        """Retrieve attribute from the screen."""
        return self.pages[pagenum].row[row-1].buf[col-1][1]

    def get_charwidth(self, pagenum, row, col):
        """Retrieve DBCS character width in bytes."""
        dbcs = self.pages[pagenum].row[row-1].double[col-1]
        if dbcs == 0:
            return 1
        elif dbcs == 1:
            return 2
        return 0

    def get_fullchar_attr(self, pagenum, row, col):
        """Retrieve SBCS or DBCS character."""
        therow = self.pages[pagenum].row[row-1]
        if therow.double[col-1] == 1:
            ca = therow.buf[col-1]
            da = therow.buf[col]
            char, attr = ca[0] + da[0], da[1]
        elif therow.double[col-1] == 0:
            ca = therow.buf[col-1]
            char, attr = ca[0], ca[1]
        else:
            char, attr = b'\0', 0
            logging.debug('DBCS buffer corrupted at %d, %d (%d)', row, col, therow.double[col-1])
        return char, attr

    def get_text_raw(self, pagenum):
        """Retrieve all raw text on a page."""
        return tuple(
            b''.join(c for c, _ in self.pages[pagenum].row[row_index].buf)
            for row_index in range(self.pages[pagenum].height)
        )

    ###########################################################################
    # logical lines

    def get_text_logical(self, pagenum, start_row, start_col, stop_row, stop_col):
        """Retrieve section of logical text for copying."""
        # include lead byte if start on trail
        if self.pages[pagenum].row[start_row-1].double[start_col-1] == 2:
            start_col -= 1
        # include trail byte if end on lead
        if self.pages[pagenum].row[stop_row-1].double[stop_col-2] == 1:
            stop_col += 1
        r, c = start_row, start_col
        full = []
        clip = []
        while r < stop_row or (r == stop_row and c < stop_col):
            clip.append(self.pages[pagenum].row[r-1].buf[c-1][0])
            c += 1
            if c > self.pages[pagenum].row[r-1].end:
                if not self.pages[pagenum].row[r-1].wrap:
                    full.append(b''.join(clip))
                    clip = []
                r += 1
                c = 1
        full.append(b''.join(clip))
        return full

    def find_start_of_line(self, pagenum, srow):
        """Find the start of the logical line that includes our current position."""
        # move up as long as previous line wraps
        while srow > 1 and self.pages[pagenum].row[srow-2].wrap:
            srow -= 1
        return srow

    def find_end_of_line(self, pagenum, srow):
        """Find the end of the logical line that includes our current position."""
        # move down as long as this line wraps
        while srow <= self.height and self.pages[pagenum].row[srow-1].wrap:
            srow += 1
        return srow

    def get_logical_line(self, pagenum, start_row, from_column=None):
        """Get bytearray of the contents of the logical line."""
        # find start of logical line
        if from_column is None:
            srow, scol = self.find_start_of_line(pagenum, start_row), 1
        else:
            srow, scol = start_row, from_column
        line = bytearray()
        # add all rows of the logical line
        for row in range(srow, self.height+1):
            therow = self.pages[pagenum].row[row-1]
            line += b''.join(pair[0] for pair in therow.buf[scol-1:therow.end])
            # continue so long as the line wraps
            if not therow.wrap:
                break
            # wrap before end of line means LF
            if therow.end < self.width:
                line += b'\n'
            # all further lines taken from start
            scol = 1
        return bytes(line)

    def get_logical_line_from(self, pagenum, srow, prompt_row, left, right):
        """Get bytearray of the contents of the logical line, adapted for INPUT."""
        # INPUT: the prompt starts at the beginning of a logical line
        # but the row may have moved up: this happens on line 24
        # in this case we need to move up to the start of the logical line
        prompt_row = self.find_start_of_line(pagenum, prompt_row)
        # find start of logical line
        srow = self.find_start_of_line(pagenum, srow)
        line = bytearray()
        # INPUT returns empty string if enter pressed below prompt row
        if srow <= prompt_row:
            # add all rows of the logical line
            for row in range(srow, self.height+1):
                therow = self.pages[pagenum].row[row-1]
                # exclude prompt, if any; only go from furthest_left to furthest_right
                if row == prompt_row:
                    rowpairs = therow.buf[:therow.end][left-1:right-1]
                else:
                    rowpairs = therow.buf[:therow.end]
                # get characters from char/attr pairs and convert to bytearray
                line += b''.join(pair[0] for pair in rowpairs)
                if not therow.wrap:
                    break
                # wrap before end of line means LF
                if therow.end < self.width:
                    line += b'\n'
        return bytes(line)



"""
PC-BASIC - signals.py
Signals for communication between interpreter and interface

(c) 2013--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""


class Event(object):
    """Signal object for input, video or audio queue."""

    def __init__(self, event_type, params=()):
        """Create signal."""
        self.event_type = event_type
        self.params = params


# audio queue signals
AUDIO_TONE = 0
AUDIO_STOP = 1
AUDIO_NOISE = 2
AUDIO_QUIT = 4
AUDIO_PERSIST = 6

# video queue signals
# save state and quit
VIDEO_QUIT = 0
# change video mode
VIDEO_SET_MODE = 1
# switch page
VIDEO_SET_PAGE = 2
# set cursor shape
VIDEO_SET_CURSOR_SHAPE = 3
# move cursor
VIDEO_MOVE_CURSOR = 5
# set cursor attribute
VIDEO_SET_CURSOR_ATTR = 6
# set border attribute
VIDEO_SET_BORDER_ATTR = 7
# put character glyph
VIDEO_PUT_GLYPH = 8
# clear rows
VIDEO_CLEAR_ROWS = 10
# scroll
VIDEO_SCROLL_UP = 11
VIDEO_SCROLL_DOWN = 12
# show/hide cursor
VIDEO_SHOW_CURSOR = 14
# set palette
VIDEO_SET_PALETTE = 15
# build glyphs
VIDEO_BUILD_GLYPHS = 16
# put pixel
VIDEO_PUT_PIXEL = 17
# put interval
VIDEO_PUT_INTERVAL = 18
VIDEO_FILL_INTERVAL = 19
# put rect
VIDEO_PUT_RECT = 20
VIDEO_FILL_RECT = 21
# copy page
VIDEO_COPY_PAGE = 28
# set caption message
VIDEO_SET_CAPTION = 29
# clipboard copy reply
VIDEO_SET_CLIPBOARD_TEXT = 30

# input queue signals
# quit interpreter
KEYB_QUIT = 0
# insert keydown
KEYB_DOWN = 5
# insert keyup
KEYB_UP = 6
# redirect or stdio input
STREAM_CHAR = 7
# redirect or stdio closed
STREAM_CLOSED = 8
# light pen events
PEN_DOWN = 101
PEN_UP = 102
PEN_MOVED = 103
# joystick events
STICK_DOWN = 201
STICK_UP = 202
STICK_MOVED = 203
# clipboard events
CLIP_COPY = 254
CLIP_PASTE = 255



"""
PC-BASIC - iostreams.py
Input/output streams

(c) 2014--2019 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

import logging
import threading
import sys
import time
import io
from contextlib import contextmanager
from collections import Iterable

from ..compat import WIN32, read_all_available
from .base import signals
from .codepage import CONTROL


# sleep period for input thread
# does not need to be very short as it reads multiple bytes in one cycle
TICK = 0.03


class IOStreams(object):
    """Manage input/output to files, printers and stdio."""

    def __init__(self, queues, codepage, input_streams, output_streams):
        """Initialise I/O streams."""
        self._queues = queues
        self._codepage = codepage
        # input; put in tuple if it's file-like so we do the right thing when looping
        if not input_streams:
            input_streams = ()
        elif hasattr(input_streams, 'read') or not isinstance(input_streams, Iterable):
            input_streams = (input_streams,)
        self._input_streams = [self._wrap_input(stream) for stream in input_streams]
        # output; put in tuple if it's file-like so we do the right thing when looping
        if not output_streams:
            output_streams = ()
        elif hasattr(output_streams, 'write') or not isinstance(output_streams, Iterable):
            output_streams = (output_streams,)
        self._output_echos = [
            self._codepage.wrap_output_stream(stream, preserve=CONTROL)
            for stream in output_streams
        ]
        # disable at start
        self._active = False
        # launch a daemon thread for input
        if self._input_streams:
            # launch a thread to allow nonblocking reads on both Windows and Unix
            thread = threading.Thread(target=self._process_input, args=())
            thread.daemon = True
            thread.start()

    def flush(self):
        """Flush output streams."""
        for f in self._output_echos:
            f.flush()

    def write(self, s):
        """Write a string/bytearray to all stream outputs."""
        for f in self._output_echos:
            f.write(s)

    def toggle_echo(self, stream):
        """Toggle copying of all screen I/O to stream."""
        if stream in self._output_echos:
            self._output_echos.remove(stream)
        else:
            self._output_echos.append(stream)

    @contextmanager
    def activate(self):
        """Grab and release input stream."""
        self._active = True
        try:
            yield
        finally:
            self._active = False

    def _wrap_input(self, stream):
        """Wrap input stream."""
        return NonBlockingInputWrapper(
            stream, self._codepage, lfcr=not WIN32 and stream.isatty()
        )

    def _process_input(self):
        """Process input from streams."""
        while True:
            time.sleep(TICK)
            if not self._active:
                continue
            queue = self._queues.inputs
            for stream in self._input_streams:
                instr = stream.read()
                if instr is None:
                    break
                elif instr:
                    queue.put(signals.Event(signals.STREAM_CHAR, (instr,)))
            else:
                # executed if not break
                continue
            # input stream is closed, remove it
            self._input_streams.remove(stream)
            # exit the interpreter if last input closed
            if not self._input_streams:
                queue.put(signals.Event(signals.STREAM_CLOSED))
                return


class NonBlockingInputWrapper(object):
    """
    Non-blocking input wrapper, converts CRLF.
    Wraps unicode or bytes stream; always produces unicode.
    """

    def __init__(self, stream, codepage, lfcr):
        """Set up codec."""
        self._stream = stream
        self._lfcr = lfcr
        # codepage, used to read unicode from bytes streams
        self._codepage = codepage

    def read(self):
        """Read all chars available; nonblocking; returns unicode."""
        # we need non-blocking readers
        s = read_all_available(self._stream)
        # can be None (closed) or b'' (no input)
        if s is None:
            return None
        elif not s:
            return u''
        if isinstance(s, bytes):
            # raw input means it's already in the BASIC codepage
            # but the keyboard functions use unicode
            # for input, don't use lead-byte buffering beyond the convert call
            s = self._codepage.str_to_unicode(s, preserve=CONTROL)
        # replace CRLF (and, if desired, LF) with CR
        s = s.replace(u'\r\n', u'\r')
        if self._lfcr:
            s = s.replace(u'\n', u'\r')
        return s
